from typing import Optional

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import *
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, lit, udf
from dasl_client.preset_development.errors import *
import uuid
from IPython import get_ipython


class PreviewParameters:
    """
    This class provides three methods for supplying input records to the preset development environment.

    **1. Input Mode:**
    In "input" mode, the user provides the schema and data directly using `StructType`, `StructField`,
    and a list of tuples representing the data. For example:

    ```python
    schema = StructType([
        StructField('name', StringType(), True),
        StructField('age', IntegerType(), True)
    ])

    data = [("Mikhail", 15), ("Zaky", 13), ("Zoya", 8)]

    data_source = PreviewParameters() \
        .from_input() \
        .set_data_schema(schema) \
        .set_data(data)
    ```

    **2. Autoloader Mode:**
    In "autoloader" mode, input is loaded using the `cloudFiles` format and settings defined in the preset's
    `autoloader` field. The format is fetched directly from the preset, while other required options must be
    provided manually. Example:

    ```python
    ds_params = PreviewParameters() \
        .from_autoloader() \
        .set_autoloader_location("s3://test-databucket/test-data") \
        .set_pretransform_name("name_of_pretransform") \
        .set_date_range("EdgeStartTimestamp", "2024-02-15 11:27:21", "2024-02-15 11:27:25")
    ```

    If you wish to skip the Silver PreTransform stage, simply omit the `pretransform_name` setting.

    **3. Table Mode:**
    This method reads input directly from a table:

    ```python
    ds_params = DataSourceParameters(spark) \
        .from_table() \
        .set_table("system.access.audit")
    ```

    **4. SilverBronze Mode:**
    "silverbronze" mode, works like a more advanced "table" mode. It allows for joining of multiple
    tables as input. This mode requires setting bronze table definitions. This mode behaves in 2
    seperate ways depending on whether an autoloader location is set or not. If an autoloader location
    is set the first entry in the bronze table definitions is used to name and alias the autoloader's
    input and these can be used in later join expressions. Used in this way, the autoloader will be
    loaded as in "autoloader" mode, and run through preBronze stages before being joined with the
    remainder of the bronze table definitions. This mimics not skipping bronze in a DataSource and
    joining what was read in silver. If an autoloader location is not set, the behaviour instead
    attempts to emulate a DataSource set to skip the bronze stage. That is, all preBronze and bronze
    stages will be skipped, and the name of the first entry in the given bronze table definitions will
    be read from instead. Any subsequent bronze table definitions will be joined against this table.

    Using no autoloader location (this will read from the first table):
    ```python
    bronze_tables = [
    {
        "name": "databricks_dev.default.sev_map",
        "alias": "tab1"
    },
    {
        "name": "databricks_dev.alan_bronze.akamai_waf",
        "alias": "tab2",
        "joinExpr": "id::string = tab2.serviceID",
        "joinType": "left"
    },
    {
        "name": "databricks_dev.alan_silver.cloudflare_hjttp_request",
        "alias": "tab3",
        "joinExpr": "tab1.id::string = tab3.ClientRequestsBytes",
        "joinType": "inner"
    }
    ]

    ds_params = (
        PreviewParameters(spark)
        .from_silverbronze_tables()
        .set_bronze_table_definitions(bronze_tables)
    )

    ps = PreviewEngine(spark, yaml_string, ds_params)
    ps.evaluate("stage.gold")
    ```

    Using an autoloader location (this will read from the autoloader and name the df tab1):
    ```python
    bronze_tables = [
    {
        "name": "tab1"
    },
    {
        "name": "databricks_dev.alan_bronze.akamai_waf",
        "alias": "tab2",
        "joinExpr": "id::string = tab2.serviceID",
        "joinType": "left"
    },
    {
        "name": "databricks_dev.alan_silver.cloudflare_hjttp_request",
        "alias": "tab3",
        "joinExpr": "tab1.id::string = tab3.ClientRequestsBytes",
        "joinType": "inner"
    }
    ]

    ds_params = (
        PreviewParameters(spark)
        .from_silverbronze_tables()
        .set_bronze_table_definitions(bronze_tables)
        .set_autoloader_location("s3://antimatter-dasl-testing/csamples3/mars/area1/")
    )
    ```

    **Note:**
    When using autoloader mode, this implementation requires a location to store a temporary schema for
    the loaded records. By default, this is set to `"dbfs:/tmp/schemas"`. You can change this using
    `set_autoloader_temp_schema_location`. Regardless of whether you use the default or a custom path,
    you must have write permissions for that location.
    """

    def __init__(self, spark: SparkSession) -> None:
        """
        Initializes the PreviewParameters instance with sparse default settings.

        Note: The preset development environment is intended to process only a small number
        of records at a time. By default, the record limit is set to 10, but this can be overridden
        if needed.

        Instance Attributes:
            mode (str): Indicates the source type ("input" or "autoloader").
            record_limit (int): Maximum number of records to load. Defaults to 10.
            autoloader_temp_schema_location (str): Temporary location to store the autoloader schema.
            time_column (str): Column name used for time-based filtering.
            start_time (str): Start time for filtering.
            end_time (str): End time for filtering.
            autoloader_location (str): Filesystem location for autoloader input.
            autoloader_format (str): Format of the data for autoloader.
            schema_file (str): Path to a file containing the schema definition.
            cloudfiles_schema_hints_file (str): Path to a file containing CloudFiles schema hints.
            cloudfiles_schema_hints (str): Directly provided CloudFiles schema hints.
            schema_uuid_str (str): Unique identifier for the schema (used in the autoloader schema path).
            schema (StructType): Schema definition for input data.
            data (dict): In-memory data used to create a DataFrame in "input" mode.
            pretransform_name (str): Name of the pre-transformation step.
            df (DataFrame): Internal Spark DataFrame loaded using the specified parameters.
        """
        self._spark = spark
        self._mode = None  # [input, table, autoloader, silverbronze]
        self._record_limit = 10
        self._autoloader_temp_schema_location = "dbfs:/tmp/schemas"
        self._gold_test_schemas = []

        self._time_column = None
        self._start_time = None
        self._end_time = None

        self._autoloader_location = None
        self._autoloader_format = None
        self._schema_file = None
        self._clouldfiles_schema_hints_file = None
        self._cloudfiles_schema_hints = None
        self._cloudfiles_reader_case_sensitive = "true"
        self._cloudfiles_multiline = "true"
        self._cloudfiles_wholetext = "false"
        self._schema_uuid_str = str(uuid.uuid4())
        self._single_variant_column = None

        self._schema = None
        self._data = None

        self._table = None

        self._bronze_tables = None

        self._pretransform_name = None

        self._df = None

    def __create_from_autoloader(self) -> DataFrame:
        stream_df = (
            self._spark.readStream.format("cloudFiles")
            .option("cloudFiles.format", self._autoloader_format)
            .option("readerCaseSensitive", self._cloudfiles_reader_case_sensitive)
        )

        # text and wholetext needs to be handled seperately.
        stream_df = (
            stream_df.option("multiline", self._cloudfiles_multiline)
            if self._autoloader_format != "text"
            else stream_df.option("wholetext", self._cloudfiles_wholetext)
        )

        if self._single_variant_column:
            stream_df = stream_df.option(
                "singleVariantColumn", self._single_variant_column
            )

        if self._schema_file:
            with open(self._schema_file, "r") as f:
                stream_df = stream_df.schema(f.read().strip())
        else:
            stream_df = (
                stream_df.option("inferSchema", "true")
                .option("cloudFiles.inferColumnTypes", "true")
                .option(
                    "cloudFiles.schemaLocation",
                    f"{self._autoloader_temp_schema_location}/{self._schema_uuid_str}",
                )
            )

        if self._cloudfiles_schema_hints:
            stream_df = stream_df.option(
                "cloudFiles.schemaHints", self._cloudfiles_schema_hints
            )
        elif self._clouldfiles_schema_hints_file:
            stream_df = stream_df.option(
                "cloudFiles.schemaHintsFile", self._clouldfiles_schema_hints_file
            )

        stream_df = stream_df.load(self._autoloader_location).limit(self._record_limit)

        query = (
            stream_df.writeStream.format("memory")
            .queryName("batch_data")
            .trigger(availableNow=True)
            .start()
        )

        query.awaitTermination()

    def __create_from_silverbronze_tables_join(self) -> DataFrame:
        if not self._bronze_tables or not len(self._bronze_tables):
            raise MissingBronzeTablesError()

        # Validate name and joinExpr are set.
        for i in range(len(self._bronze_tables)):
            if not self._bronze_tables[i].get("name", None):
                raise MissingBronzeTableFieldError("name")
            if i > 0 and not self._bronze_tables[i].get("joinExpr", None):
                raise MissingBronzeTableFieldError("joinExpr")

        # If there is an autoloader location given, we create the df now and
        # then allow preBronze stage to run. Otherwise we skip preBronze stages
        # and as part of running the silverbronze joins we create the df from
        # the first entry in the bronze tables list.
        df = None
        if self._autoloader_location:
            self.__create_from_autoloader()
            df = self._spark.table("batch_data").alias(
                self._bronze_tables[0].get("name", "")
            )  # Use first's name.

        return df

    def __enter__(self):
        """
        Creates a DataFrame with data using the method specified. In the case of "autoloader",
        this will stream to a DataFrame that is then treated as a batch. This allows for easier
        emulation of some operations, while not giving up some of the options allowed by
        streaming.

        Returns:
            DataFrame: The resulting DataFrame with input data.
        """
        if self._mode == "input":
            self._df = self._spark.createDataFrame(self._data, self._schema)
        elif self._mode == "table":
            self._df = self._spark.table(self._table).limit(self._record_limit)
        elif self._mode == "autoloader":
            self.__create_from_autoloader()
            self._df = self._spark.table("batch_data")
        elif self._mode == "silverbronze":
            self._df = self.__create_from_silverbronze_tables_join()

        return self._df

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Cleans up the temporary schema created for streaming mode, if it was created.
        """

        # Get the Databricks built-in functions out the namespace.
        ipython = get_ipython()
        if ipython is not None:
            dbutils = ipython.user_ns["dbutils"]

            dbutils.fs.rm(
                f"{self._autoloader_temp_schema_location}/{self._schema_uuid_str}",
                recurse=True,
            )
            for gold_test_schema in self._gold_test_schemas:
                dbutils.fs.rm(
                    f"{self._autoloader_temp_schema_location}/{gold_test_schema}",
                    recurse=True,
                )
        else:
            leaked_lines = [
                f"FYI, we are leaking temp data {self._autoloader_temp_schema_location}/{self._schema_uuid_str}",
                *[
                    f"{self._autoloader_temp_schema_location}/{x}"
                    for x in self._gold_test_schemas
                ],
            ]
            print(", ".join(leaked_lines))
        self._gold_test_schemas = []

    def from_input(self):
        """
        Set the data source loader to "input" mode. Requires a schema and data to be provided.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._mode = "input"
        return self

    def from_autoloader(self):
        """
        Set the data source loader to "autoloader" mode. Requires at least autoloader location
        to be provided.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._mode = "autoloader"
        return self

    def from_table(self):
        """
        Set the data source loader to "table" mode. Requires a table name to be provided.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._mode = "table"
        return self

    def from_silverbronze_tables(self):
        """
        Set the data source loader to "bronze tables" mode. Requires a list of bronze table
        definitions to be provided.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._mode = "silverbronze"
        return self

    def set_bronze_table_definitions(self, definitions: List[Dict[str, str]]):
        """
        Set the bronze table definitions for bronze tables mode. `name` and `joinExpr` are
        required. If `alias` is not provided, one can use the `name` to refer to the table.
        If `joinType` is not provided, "left" is used as a default value. If pr

        [
            {
                "name": "name",
                "alias": "alias1",
                "joinType": "inner",
                "joinExpr": "base_table.col1 = alias1.col1
            },
            ...
        ]
        """
        self._bronze_tables = definitions
        return self

    def set_autoloader_temp_schema_location(self, path: str):
        """
        Set the location for the autoloader's streaming mode schema to be created. This is
        deleted at the end of a run.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._autoloader_temp_schema_location = path
        return self

    def get_autoloader_temp_schema_location(self) -> str:
        """
        Get the location for the autoloader's streaming mode schema to be created.

        Returns:
             str: The location for the autoloader's streaming mode schema to be created.
        """
        return self._autoloader_temp_schema_location

    def set_data_schema(self, schema: StructType):
        """
        Set the input schema for "input" mode. For example:

        StructType([
            StructField('name', StringType(), True),
            StructField('age', IntegerType(), True)
        ])

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._schema = schema
        return self

    def set_data(self, data: Dict[str, str]):
        """
        Set the input data for "input" mode. For example:

        [("Peter", 15), ("Urvi", 13), ("Graeme", 8)]

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._data = data
        return self

    def set_autoloader_location(self, location: str):
        """
        Set where to load data from for "autoloader" mode.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._autoloader_location = location
        return self

    def _set_autoloader_format(self, file_format: str):
        """
        Used internally to set the autoloader format.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        if file_format.lower() == "jsonl":
            self._autoloader_format = "json"
            self._cloudfiles_multiline = "false"
            return self
        if file_format.lower() == "wholetext":
            self._autoloader_format = "text"
            self._cloudfiles_wholetext = "true"
            return self
        self._autoloader_format = file_format
        return self

    def _set_autoloader_schema_file(self, path: str):
        """
        Set the schema file path for "autoloader" mode.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._schema_file = path
        return self

    def _set_autoloader_cloudfiles_schema_hint_file(self, path: str):
        """
        Set the cloudFiles schema hints file path for "autoloader" mode.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._clouldfiles_schema_hints_file = path
        return self

    def _set_autoloader_cloudfiles_schema_hints(self, cloudfiles_schema_hints: str):
        """
        Set the cloudFiles schema hints string for "autoloader" mode.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._cloudfiles_schema_hints = cloudfiles_schema_hints
        return self

    def set_pretransform_name(self, pretransform_name: str):
        """
        Set the pretransform name to use, if desired. If not set, Silver PreTransform
        will be skipped.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._pretransform_name = pretransform_name
        return self

    def set_date_range(self, column: str, start_time: str, end_time: str):
        """
        Set the TIMESTAMP column and date range to use as the input data filter to
        limit the number of records retrieved by the loader.

        Both start and end time must be TIMESTAMP compatible.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._time_column = column
        self._start_time = start_time
        self._end_time = end_time
        return self

    def set_input_record_limit(self, record_limit: int):
        """
        Set the LIMIT clause when retrieving records from the data source.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._record_limit = record_limit
        return self

    def set_table(self, table_name: str):
        """
        Set Unity Catalog table name for "table" mode.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._table = table_name
        return self

    def _set_load_as_single_variant(self, col_name: Optional[str] = None):
        """
        Enable loadAsSingleVariant mode. This will ingest data into a single VARIANT-typed column.
        The default name of that column is `data`.

        Returns:
            PreviewParameters: The current instance with updated configuration.
        """
        self._single_variant_column = col_name if col_name is not None else "data"
        return self

    def add_gold_schema_table(self, gold_schema_table_name: str):
        """
        Add a gold schema temporary table name that will need to be cleaned
        up at the end of the run.
        """
        self._gold_test_schemas.append(gold_schema_table_name)



from pydantic import BaseModel
from typing import Dict, List, Optional

from dasl_api import (
    ContentV1DatasourcePresetAutoloaderCloudFiles,
    CoreV1DataSource,
    CoreV1DataSourceAutoloaderSpec,
    CoreV1DataSourceSpec,
    CoreV1DataSourceSpecCustom,
    CoreV1DataSourceSpecGold,
    CoreV1DataSourceSpecGoldPresetOverrides,
    CoreV1DataSourceSpecGoldPresetOverridesAddTablesInner,
    CoreV1DataSourceSpecGoldPresetOverridesAddTablesInnerCustom,
    CoreV1DataSourceSpecGoldPresetOverridesModifyTablesInner,
    CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInnerCustom,
    CoreV1DataSourceSpecSilver,
    CoreV1DataSourceSpecSilverBronzeTablesInner,
    CoreV1DataSourceSpecSilverBronzeTablesInnerWatermark,
    CoreV1DataSourceSpecSilverPreTransform,
    CoreV1DataSourceSpecSilverPreTransformCustom,
    CoreV1DataSourceSpecSilverPreTransformPresetOverrides,
    CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInner,
    CoreV1DataSourceSpecSilverTransformPresetOverridesAddTablesInner,
    CoreV1DataSourceSpecSilverTransform,
    CoreV1DataSourceSpecSilverTransformPresetOverrides,
    CoreV1DataSourceSpecBronze,
    CoreV1DataSourceSpecBronzeClustering,
    CoreV1DataSourceFieldSpec,
    CoreV1DataSourceFieldSpecAssertInner,
    CoreV1DataSourceFieldSpecJoin,
    CoreV1DataSourceFieldSpecJoinWithCSV,
    CoreV1DataSourceFieldUtils,
    CoreV1DataSourceFieldUtilsUnreferencedColumns,
    CoreV1DataSourceFieldUtilsJsonExtractInner,
    CoreV1DataSourcePrimaryKeySpec,
)

from .helpers import Helpers
from .types import Metadata, ResourceStatus, Schedule


class FieldSpec(BaseModel):
    """
    FieldSpec

    Attributes:
        name (Optional[str]):
            The name of the field.
        comment (Optional[str]):
            The comment to apply to the field.
        var_assert (Optional[List[FieldSpec.Assert]]):
            A list of SQL expressions that must evaluate to true for every
            processed row. If the assertion is false, an operational alert
            is raised using 'message' for each row.
        var_from (Optional[str]):
            This field obtains its value from the source column of this name.
            Use this to bring in a column from some upstream table.
        alias (Optional[str]):
            This field obtains its value from the destination (transformed)
            column of this name. Use this to alias a column from within the
            same table (ie. silver table). You cannot alias a column from
            some upstream table.
        expr (Optional[str]):
            This field obtains its value from the given SQL expression.
        literal (Optional[str]):
            This field obtains its value from the given literal string. For
            other data types, use expr.
        join (Optional[FieldSpec.Join]):
            This field obtains its value from joining to another table.
    """

    class Assert(BaseModel):
        """
        An assertion within a FieldSpec.

        Attributes:
            expr (Optional[str]):
                The SQL expression that must evaluate to true for every
                processed row.
            message (Optional[str]):
                The message to include in the operational alert if the
                assertion fails.
        """

        expr: Optional[str] = None
        message: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceFieldSpecAssertInner],
        ) -> "FieldSpec.Assert":
            if obj is None:
                return None
            return FieldSpec.Assert(
                expr=obj.expr,
                message=obj.message,
            )

        def to_api_obj(self) -> CoreV1DataSourceFieldSpecAssertInner:
            return CoreV1DataSourceFieldSpecAssertInner(
                expr=self.expr,
                message=self.message,
            )

    class Join(BaseModel):
        """
        A join expression within a FieldSpec.

        Attributes:
            with_table (Optional[str]):
                The table to join to.
            with_csv (Optional[FieldSpec.Join.WithCSV]):
                The CSV configuration used for the join.
            lhs (Optional[str]):
                The column in the source dataframe to join on.
            rhs (Optional[str]):
                The column in the joined table to join on.
            select (Optional[str]):
                A SQL expression to create the new field from the joined
                dataset.
        """

        class WithCSV(BaseModel):
            """
            A CSV file used for joins within a FieldSpec.

            Attributes:
                path (Optional[str]):
                    The path to the CSV file.
            """

            path: Optional[str] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1DataSourceFieldSpecJoinWithCSV],
            ) -> "FieldSpec.Join.WithCSV":
                if obj is None:
                    return None
                return FieldSpec.Join.WithCSV(path=obj.path)

            def to_api_obj(self) -> CoreV1DataSourceFieldSpecJoinWithCSV:
                return CoreV1DataSourceFieldSpecJoinWithCSV(path=self.path)

        with_table: Optional[str] = None
        with_csv: Optional["FieldSpec.Join.WithCSV"] = None
        lhs: Optional[str] = None
        rhs: Optional[str] = None
        select: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceFieldSpecJoin],
        ) -> "FieldSpec.Join":
            if obj is None:
                return None
            return FieldSpec.Join(
                with_table=obj.with_table,
                with_csv=FieldSpec.Join.WithCSV.from_api_obj(obj.with_csv),
                lhs=obj.lhs,
                rhs=obj.rhs,
                select=obj.select,
            )

        def to_api_obj(self) -> CoreV1DataSourceFieldSpecJoin:
            return CoreV1DataSourceFieldSpecJoin(
                with_table=self.with_table,
                with_csv=Helpers.maybe(lambda o: o.to_api_obj(), self.with_csv),
                lhs=self.lhs,
                rhs=self.rhs,
                select=self.select,
            )

    name: Optional[str] = None
    comment: Optional[str] = None
    var_assert: Optional[List["FieldSpec.Assert"]] = None
    var_from: Optional[str] = None
    alias: Optional[str] = None
    expr: Optional[str] = None
    literal: Optional[str] = None
    join: Optional["FieldSpec.Join"] = None

    @staticmethod
    def from_api_obj(obj: Optional[CoreV1DataSourceFieldSpec]) -> "FieldSpec":
        if obj is None:
            return None
        var_assert = None
        if obj.var_assert is not None:
            var_assert = [
                FieldSpec.Assert.from_api_obj(item) for item in obj.var_assert
            ]
        return FieldSpec(
            name=obj.name,
            comment=obj.comment,
            var_assert=var_assert,
            var_from=obj.var_from,
            alias=obj.alias,
            expr=obj.expr,
            literal=obj.literal,
            join=FieldSpec.Join.from_api_obj(obj.join),
        )

    def to_api_obj(self) -> CoreV1DataSourceFieldSpec:
        var_assert = None
        if self.var_assert is not None:
            var_assert = [item.to_api_obj() for item in self.var_assert]
        to_api_obj = lambda o: o.to_api_obj()
        return CoreV1DataSourceFieldSpec(
            name=self.name,
            comment=self.comment,
            var_assert=var_assert,
            var_from=self.var_from,
            alias=self.alias,
            expr=self.expr,
            literal=self.literal,
            join=Helpers.maybe(to_api_obj, self.join),
        )


class FieldUtils(BaseModel):
    """
    FieldUtils

    Attributes:
        unreferenced_columns (Optional[FieldUtils.UnreferencedColumns]):
            Defines whether columns not referenced in the FieldSpecs should
            be preserved or omitted.
        json_extract (Optional[List[FieldUtils.JsonExtract]]):
            A list of configurations for extracting JSON fields from a column.
    """

    class UnreferencedColumns(BaseModel):
        """
        Configuration related to unreferenced columns.

        Attributes:
            preserve (Optional[bool]):
                Indicates whether columns not referenced in the FieldSpecs
                should be preserved.
            embed_column (Optional[str]):
                Specifies a name for a new column to contain all unreferenced
                fields.
            omit_columns (Optional[List[str]]):
                Lists columns to exclude from the output.
            duplicate_prefix (Optional[str]):
                Adds a prefix to resolve ambiguous duplicate field names.
        """

        preserve: Optional[bool] = None
        embed_column: Optional[str] = None
        omit_columns: Optional[List[str]] = None
        duplicate_prefix: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceFieldUtilsUnreferencedColumns],
        ) -> "FieldUtils.UnreferencedColumns":
            if obj is None:
                return None
            return FieldUtils.UnreferencedColumns(
                preserve=obj.preserve,
                embed_column=obj.embed_column,
                omit_columns=obj.omit_columns,
                duplicate_prefix=obj.duplicate_prefix,
            )

        def to_api_obj(self) -> CoreV1DataSourceFieldUtilsUnreferencedColumns:
            return CoreV1DataSourceFieldUtilsUnreferencedColumns(
                preserve=self.preserve,
                embed_column=self.embed_column,
                omit_columns=self.omit_columns,
                duplicate_prefix=self.duplicate_prefix,
            )

    class JsonExtract(BaseModel):
        """
        Configuration for extracting JSON fields from table columns.

        Attributes:
            source (Optional[str]):
                The column name containing JSON string(s) to extract from.
            omit_fields (Optional[List[str]]):
                Specifies high-level fields to exclude from extraction.
            duplicate_prefix (Optional[str]):
                Adds a prefix to resolve duplicate field names during
                extraction.
            embed_column (Optional[str]):
                Specifies a column name to store the extracted JSON object.
        """

        source: Optional[str] = None
        omit_fields: Optional[List[str]] = None
        duplicate_prefix: Optional[str] = None
        embed_column: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceFieldUtilsJsonExtractInner],
        ) -> "FieldUtils.JsonExtract":
            if obj is None:
                return None
            return FieldUtils.JsonExtract(
                source=obj.source,
                omit_fields=obj.omit_fields,
                duplicate_prefix=obj.duplicate_prefix,
                embed_column=obj.embed_column,
            )

        def to_api_obj(self) -> CoreV1DataSourceFieldUtilsJsonExtractInner:
            return CoreV1DataSourceFieldUtilsJsonExtractInner(
                source=self.source,
                omit_fields=self.omit_fields,
                duplicate_prefix=self.duplicate_prefix,
                embed_column=self.embed_column,
            )

    unreferenced_columns: Optional["FieldUtils.UnreferencedColumns"] = None
    json_extract: Optional[List["FieldUtils.JsonExtract"]] = None

    @staticmethod
    def from_api_obj(obj: Optional[CoreV1DataSourceFieldUtils]) -> "FieldUtils":
        if obj is None:
            return None
        json_extract = None
        if obj.json_extract is not None:
            json_extract = [
                FieldUtils.JsonExtract.from_api_obj(item) for item in obj.json_extract
            ]
        return FieldUtils(
            unreferenced_columns=FieldUtils.UnreferencedColumns.from_api_obj(
                obj.unreferenced_columns
            ),
            json_extract=json_extract,
        )

    def to_api_obj(self) -> CoreV1DataSourceFieldUtils:
        json_extract = None
        if self.json_extract is not None:
            json_extract = [item.to_api_obj() for item in self.json_extract]
        return CoreV1DataSourceFieldUtils(
            unreferenced_columns=self.unreferenced_columns.to_api_obj(),
            json_extract=json_extract,
        )


class BronzeSpec(BaseModel):
    """
    Configuration for bronze table within a DataSource.

    Attributes:
        clustering (Optional[BronzeSpec.Clustering]):
            Describes optional liquid clustering configuration for the
            bronze table.
        bronze_table (Optional[str]):
            The name of the bronze table to create and hold the imported data.
        skip_bronze_loading (Optional[bool]):
            Indicates whether to skip the bronze loading step.
        load_as_single_variant (Optional[bool]):
            Indicates whether to ingest data into a single VARIANT-typed column called `data`
        pre_transform (Optional[List[List[str]]]):
            A list of pre-transform steps to execute.
            The outer list form stages and the inner list contains SQL select expressions to be executed within each stage
    """

    class Clustering(BaseModel):
        """
        Configuration of liquid clustering for a bronze table.

        Attributes:
            column_names (Optional[List[str]]):
                List of column names to include in liquid clustering.
            time_column (Optional[str]):
                Name of the column that holds 'time' information for
                clustering.
        """

        column_names: Optional[List[str]] = None
        time_column: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceSpecBronzeClustering],
        ) -> "BronzeSpec.Clustering":
            if obj is None:
                return None
            return BronzeSpec.Clustering(
                column_names=obj.column_names,
                time_column=obj.time_column,
            )

        def to_api_obj(self) -> CoreV1DataSourceSpecBronzeClustering:
            return CoreV1DataSourceSpecBronzeClustering(
                column_names=self.column_names,
                time_column=self.time_column,
            )

    clustering: Optional["BronzeSpec.Clustering"] = None
    bronze_table: Optional[str] = None
    skip_bronze_loading: Optional[bool] = None
    load_as_single_variant: Optional[bool] = None
    pre_transform: Optional[List[List[str]]] = None

    @staticmethod
    def from_api_obj(obj: Optional[CoreV1DataSourceSpecBronze]) -> "BronzeSpec":
        if obj is None:
            return None
        return BronzeSpec(
            clustering=BronzeSpec.Clustering.from_api_obj(obj.clustering),
            bronze_table=obj.bronze_table,
            skip_bronze_loading=obj.skip_bronze_loading,
            load_as_single_variant=obj.load_as_single_variant,
            pre_transform=obj.pre_transform,
        )

    def to_api_obj(self) -> CoreV1DataSourceSpecBronze:
        return CoreV1DataSourceSpecBronze(
            clustering=Helpers.maybe(lambda o: o.to_api_obj(), self.clustering),
            bronze_table=self.bronze_table,
            skip_bronze_loading=self.skip_bronze_loading,
            load_as_single_variant=self.load_as_single_variant,
            pre_transform=self.pre_transform,
        )


class SilverSpec(BaseModel):
    """
    Configuration for silver table in a DataSource.

    Attributes:
        bronze_tables (Optional[List[SilverSpec.BronzeTable]]):
            A list of bronze tables to be joined for silver transformation.
        pre_transform (Optional[SilverSpec.PreTransform]):
            Pretransformation configuration.
        transform (Optional[SilverSpec.Transform]):
            Transformation configuration for silver processing.
    """

    class BronzeTable(BaseModel):
        """
        Reference to a bronze table for a silver table.

        Attributes:
            name (Optional[str]):
                Name of the bronze table.
            streaming (Optional[bool]):
                True if the input should be streamed from the bronze table.
            watermark (Optional[SilverSpec.BronzeTable.Watermark]):
                Bronze table watermark.
            alias (Optional[str]):
                Alias name for the table.
            join_type (Optional[str]):
                How to join to the preceding table.
            join_expr (Optional[str]):
                The join condition expression.
        """

        class Watermark(BaseModel):
            """
            Watermark for a bronze source table within a silver table.

            Attributes:
                event_time_column (Optional[str]):
                    Which column is the event time for the delay threshold.
                delay_threshold (Optional[str]):
                    A time duration string for the watermark delay.
                drop_duplicates (Optional[List[str]]):
                    Columns to pass to pyspark dropDuplicates.
            """

            event_time_column: Optional[str] = None
            delay_threshold: Optional[str] = None
            drop_duplicates: Optional[List[str]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1DataSourceSpecSilverBronzeTablesInnerWatermark],
            ) -> "SilverSpec.BronzeTable.Watermark":
                if obj is None:
                    return None
                return SilverSpec.BronzeTable.Watermark(
                    event_time_column=obj.event_time_column,
                    delay_threshold=obj.delay_threshold,
                    drop_duplicates=obj.drop_duplicates,
                )

            def to_api_obj(
                self,
            ) -> CoreV1DataSourceSpecSilverBronzeTablesInnerWatermark:
                return CoreV1DataSourceSpecSilverBronzeTablesInnerWatermark(
                    event_time_column=self.event_time_column,
                    delay_threshold=self.delay_threshold,
                    drop_duplicates=self.drop_duplicates,
                )

        name: Optional[str] = None
        streaming: Optional[bool] = None
        watermark: Optional["SilverSpec.BronzeTable.Watermark"] = None
        alias: Optional[str] = None
        join_type: Optional[str] = None
        join_expr: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceSpecSilverBronzeTablesInner],
        ) -> "SilverSpec.BronzeTable":
            if obj is None:
                return None
            return SilverSpec.BronzeTable(
                name=obj.name,
                streaming=obj.streaming,
                watermark=SilverSpec.BronzeTable.Watermark.from_api_obj(obj.watermark),
                alias=obj.alias,
                join_type=obj.join_type,
                join_expr=obj.join_expr,
            )

        def to_api_obj(self) -> CoreV1DataSourceSpecSilverBronzeTablesInner:
            return CoreV1DataSourceSpecSilverBronzeTablesInner(
                name=self.name,
                streaming=self.streaming,
                watermark=Helpers.maybe(lambda o: o.to_api_obj(), self.watermark),
                alias=self.alias,
                join_type=self.join_type,
                join_expr=self.join_expr,
            )

    class PreTransform(BaseModel):
        """
        Pre-transform for a silver table.

        Attributes:
            use_preset (Optional[str]):
                Preset to use.
            skip_pre_transform (Optional[bool]):
                If True, skip pre-transform entirely.
            custom (Optional[SilverSpec.PreTransform.Custom]):
                Custom pretransform function and options.
            filter (Optional[str]):
                A SQL filter to apply at the beginning of the preTransform
                phase.
            post_filter (Optional[str]):
                A SQL filter to apply at the end of the preTransform phase.
            preset_overrides (Optional[SilverSpec.PreTransform.PresetOverrides]):
                Overrides for preset filters.
            add_fields (Optional[List[FieldSpec]]):
                User defined fields to add to the transformation.
        """

        class Custom(BaseModel):
            """
            Custom pre-transform function for silver table.

            Attributes:
                function (Optional[str]):
                options (Optional[Dict[str, str]]):
            """

            function: Optional[str] = None
            options: Optional[Dict[str, str]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1DataSourceSpecSilverPreTransformCustom],
            ) -> "SilverSpec.PreTransform.Custom":
                if obj is None:
                    return None
                return SilverSpec.PreTransform.Custom(
                    function=obj.function,
                    options=obj.options,
                )

            def to_api_obj(self) -> CoreV1DataSourceSpecSilverPreTransformCustom:
                return CoreV1DataSourceSpecSilverPreTransformCustom(
                    function=self.function,
                    options=self.options,
                )

        class PresetOverrides(BaseModel):
            """
            Overrides for the preset.

            Attributes:
                omit_fields (Optional[List[str]]):
                    A list of fields to omit from the chosen preset.
            """

            omit_fields: Optional[List[str]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1DataSourceSpecSilverPreTransformPresetOverrides],
            ) -> "SilverSpec.PreTransform.PresetOverrides":
                if obj is None:
                    return None
                return SilverSpec.PreTransform.PresetOverrides(
                    omit_fields=obj.omit_fields
                )

            def to_api_obj(
                self,
            ) -> CoreV1DataSourceSpecSilverPreTransformPresetOverrides:
                return CoreV1DataSourceSpecSilverPreTransformPresetOverrides(
                    omit_fields=self.omit_fields,
                )

        use_preset: Optional[str] = None
        skip_pre_transform: Optional[bool] = None
        custom: Optional["SilverSpec.PreTransform.Custom"] = None
        filter: Optional[str] = None
        post_filter: Optional[str] = None
        preset_overrides: Optional["SilverSpec.PreTransform.PresetOverrides"] = None
        add_fields: Optional[List[FieldSpec]] = None
        utils: Optional[FieldUtils] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceSpecSilverPreTransform],
        ) -> "SilverSpec.PreTransform":
            if obj is None:
                return None
            add_fields = None
            if obj.add_fields is not None:
                add_fields = [FieldSpec.from_api_obj(item) for item in obj.add_fields]
            return SilverSpec.PreTransform(
                use_preset=obj.use_preset,
                skip_pre_transform=obj.skip_pre_transform,
                custom=SilverSpec.PreTransform.Custom.from_api_obj(obj.custom),
                filter=obj.filter,
                post_filter=obj.post_filter,
                preset_overrides=SilverSpec.PreTransform.PresetOverrides.from_api_obj(
                    obj.preset_overrides
                ),
                add_fields=add_fields,
                utils=FieldUtils.from_api_obj(obj.utils),
            )

        def to_api_obj(self) -> CoreV1DataSourceSpecSilverPreTransform:
            add_fields = None
            if self.add_fields is not None:
                add_fields = [item.to_api_obj() for item in self.add_fields]
            to_api_obj = lambda o: o.to_api_obj()
            return CoreV1DataSourceSpecSilverPreTransform(
                use_preset=self.use_preset,
                skip_pre_transform=self.skip_pre_transform,
                custom=Helpers.maybe(to_api_obj, self.custom),
                filter=self.filter,
                post_filter=self.post_filter,
                preset_overrides=Helpers.maybe(to_api_obj, self.preset_overrides),
                add_fields=add_fields,
                utils=Helpers.maybe(to_api_obj, self.utils),
            )

    class Transform(BaseModel):
        """
        Silver table transform.

        Attributes:
            skip_silver_transform (Optional[bool]):
                If True, skip transform entirely.
            preset_overrides (Optional[SilverSpec.Transform.PresetOverrides]):
                Preset overrides for the silver transformation.
        """

        class PresetOverrides(BaseModel):
            """
            Overrides for preset transform settings.

            Attributes:
                modify_tables (Optional[List[SilverSpec.Transform.PresetOverrides.ModifyTables]]):
                    Modifications forexisting tables.
                omit_tables (Optional[List[str]]):
                    A list of tables to omit from the preset.
                add_tables (Optional[List[SilverSpec.Transform.PresetOverrides.AddTables]]):
                    User defined tables to include in the transformation.
            """

            class Custom(BaseModel):
                """
                Custom function for use in silver table transform.

                Attributes:
                    function (Optional[str]):
                    options (Optional[Dict[str, str]]):
                """

                function: Optional[str] = None
                options: Optional[Dict[str, str]] = None

                @staticmethod
                def from_api_obj(
                    obj: Optional[
                        CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInnerCustom
                    ],
                ) -> "SilverSpec.Transform.PresetOverrides.Custom":
                    if obj is None:
                        return None
                    return SilverSpec.Transform.PresetOverrides.Custom(
                        function=obj.function,
                        options=obj.options,
                    )

                def to_api_obj(
                    self,
                ) -> CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInnerCustom:
                    return CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInnerCustom(
                        function=self.function,
                        options=self.options,
                    )

            class ModifyTables(BaseModel):
                """
                Table modifications as part of a silver transform.

                Attributes:
                    name (Optional[str]):
                    custom (Optional[SilverSpec.Transform.PresetOverrides.Custom]):
                    omit_fields (Optional[List[str]]):
                    override_liquid_columns (Optional[List[str]]):
                    add_fields (Optional[List[FieldSpec]]):
                    filter (Optional[str]):
                    post_filter (Optional[str]):
                    utils (Optional[FieldUtils]):
                """

                name: Optional[str] = None
                custom: Optional["SilverSpec.Transform.PresetOverrides.Custom"] = None
                omit_fields: Optional[List[str]] = None
                override_liquid_columns: Optional[List[str]] = None
                add_fields: Optional[List[FieldSpec]] = None
                filter: Optional[str] = None
                post_filter: Optional[str] = None
                utils: Optional[FieldUtils] = None

                @staticmethod
                def from_api_obj(
                    obj: Optional[
                        CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInner
                    ],
                ) -> "SilverSpec.Transform.PresetOverrides.ModifyTables":
                    if obj is None:
                        return None
                    add_fields = None
                    if obj.add_fields is not None:
                        add_fields = [
                            FieldSpec.from_api_obj(item) for item in obj.add_fields
                        ]
                    return SilverSpec.Transform.PresetOverrides.ModifyTables(
                        name=obj.name,
                        custom=SilverSpec.Transform.PresetOverrides.Custom.from_api_obj(
                            obj.custom
                        ),
                        omit_fields=obj.omit_fields,
                        override_liquid_columns=obj.override_liquid_columns,
                        add_fields=add_fields,
                        filter=obj.filter,
                        post_filter=obj.post_filter,
                        utils=FieldUtils.from_api_obj(obj.utils),
                    )

                def to_api_obj(
                    self,
                ) -> (
                    CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInner
                ):
                    add_fields = None
                    if self.add_fields is not None:
                        add_fields = [item.to_api_obj() for item in self.add_fields]
                    to_api_obj = lambda o: o.to_api_obj()
                    return CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInner(
                        name=self.name,
                        custom=Helpers.maybe(to_api_obj, self.custom),
                        omit_fields=self.omit_fields,
                        override_liquid_columns=self.override_liquid_columns,
                        add_fields=add_fields,
                        filter=self.filter,
                        post_filter=self.post_filter,
                        utils=Helpers.maybe(to_api_obj, self.utils),
                    )

            class AddTables(BaseModel):
                """
                Tables to add during a silver table transform.

                Attributes:
                    custom (Optional[SilverSpec.Transform.PresetOverrides.Custom]):
                    name (Optional[str]):
                    filter (Optional[str]):
                    post_filter (Optional[str]):
                    override_liquid_columns (Optional[List[str]]):
                    fields (Optional[List[FieldSpec]]):
                    utils (Optional[FieldUtils]):
                """

                custom: Optional["SilverSpec.Transform.PresetOverrides.Custom"] = None
                name: Optional[str] = None
                filter: Optional[str] = None
                post_filter: Optional[str] = None
                override_liquid_columns: Optional[List[str]] = None
                fields: Optional[List[FieldSpec]] = None
                utils: Optional[FieldUtils] = None

                @staticmethod
                def from_api_obj(
                    obj: Optional[
                        CoreV1DataSourceSpecSilverTransformPresetOverridesAddTablesInner
                    ],
                ) -> "SilverSpec.Transform.PresetOverrides.AddTables":
                    if obj is None:
                        return None
                    fields = None
                    if obj.fields is not None:
                        fields = [FieldSpec.from_api_obj(item) for item in obj.fields]
                    return SilverSpec.Transform.PresetOverrides.AddTables(
                        custom=SilverSpec.Transform.PresetOverrides.Custom.from_api_obj(
                            obj.custom
                        ),
                        name=obj.name,
                        filter=obj.filter,
                        post_filter=obj.post_filter,
                        override_liquid_columns=obj.override_liquid_columns,
                        fields=fields,
                        utils=FieldUtils.from_api_obj(obj.utils),
                    )

                def to_api_obj(
                    self,
                ) -> CoreV1DataSourceSpecSilverTransformPresetOverridesAddTablesInner:
                    fields = None
                    if self.fields is not None:
                        fields = [item.to_api_obj() for item in self.fields]
                    to_api_obj = lambda o: o.to_api_obj()
                    return CoreV1DataSourceSpecSilverTransformPresetOverridesAddTablesInner(
                        custom=Helpers.maybe(to_api_obj, self.custom),
                        name=self.name,
                        filter=self.filter,
                        post_filter=self.post_filter,
                        override_liquid_columns=self.override_liquid_columns,
                        fields=fields,
                        utils=Helpers.maybe(to_api_obj, self.utils),
                    )

            modify_tables: Optional[
                List["SilverSpec.Transform.PresetOverrides.ModifyTables"]
            ] = None
            omit_tables: Optional[List[str]] = None
            add_tables: Optional[
                List["SilverSpec.Transform.PresetOverrides.AddTables"]
            ] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1DataSourceSpecSilverTransformPresetOverrides],
            ) -> "SilverSpec.Transform.PresetOverrides":
                if obj is None:
                    return None
                modify_tables = None
                if obj.modify_tables is not None:
                    modify_tables = [
                        SilverSpec.Transform.PresetOverrides.ModifyTables.from_api_obj(
                            item
                        )
                        for item in obj.modify_tables
                    ]
                add_tables = None
                if obj.add_tables is not None:
                    add_tables = [
                        SilverSpec.Transform.PresetOverrides.AddTables.from_api_obj(
                            item
                        )
                        for item in obj.add_tables
                    ]
                return SilverSpec.Transform.PresetOverrides(
                    modify_tables=modify_tables,
                    omit_tables=obj.omit_tables,
                    add_tables=add_tables,
                )

            def to_api_obj(self) -> CoreV1DataSourceSpecSilverTransformPresetOverrides:
                modify_tables = None
                if self.modify_tables is not None:
                    modify_tables = [item.to_api_obj() for item in self.modify_tables]
                add_tables = None
                if self.add_tables is not None:
                    add_tables = [item.to_api_obj() for item in self.add_tables]
                return CoreV1DataSourceSpecSilverTransformPresetOverrides(
                    modify_tables=modify_tables,
                    omit_tables=self.omit_tables,
                    add_tables=add_tables,
                )

        skip_silver_transform: Optional[bool] = None
        do_not_materialize: Optional[bool] = None
        preset_overrides: Optional["SilverSpec.Transform.PresetOverrides"] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceSpecSilverTransform],
        ) -> "SilverSpec.Transform":
            if obj is None:
                return None
            return SilverSpec.Transform(
                skip_silver_transform=obj.skip_silver_transform,
                do_not_materialize=obj.do_not_materialize,
                preset_overrides=SilverSpec.Transform.PresetOverrides.from_api_obj(
                    obj.preset_overrides
                ),
            )

        def to_api_obj(self) -> CoreV1DataSourceSpecSilverTransform:
            return CoreV1DataSourceSpecSilverTransform(
                skipSilverTransform=self.skip_silver_transform,
                doNotMaterialize=self.do_not_materialize,
                presetOverrides=Helpers.maybe(
                    lambda o: o.to_api_obj(), self.preset_overrides
                ),
            )

    bronze_tables: Optional[List["SilverSpec.BronzeTable"]] = None
    pre_transform: Optional["SilverSpec.PreTransform"] = None
    transform: Optional["SilverSpec.Transform"] = None

    @staticmethod
    def from_api_obj(obj: Optional[CoreV1DataSourceSpecSilver]) -> "SilverSpec":
        if obj is None:
            return None
        bronze_tables = None
        if obj.bronze_tables is not None:
            bronze_tables = [
                SilverSpec.BronzeTable.from_api_obj(item) for item in obj.bronze_tables
            ]
        return SilverSpec(
            bronze_tables=bronze_tables,
            pre_transform=SilverSpec.PreTransform.from_api_obj(obj.pre_transform),
            transform=SilverSpec.Transform.from_api_obj(obj.transform),
        )

    def to_api_obj(self) -> CoreV1DataSourceSpecSilver:
        bronze_tables = None
        if self.bronze_tables is not None:
            bronze_tables = [item.to_api_obj() for item in self.bronze_tables]
        to_api_obj = lambda o: o.to_api_obj()
        return CoreV1DataSourceSpecSilver(
            bronze_tables=bronze_tables,
            pre_transform=Helpers.maybe(to_api_obj, self.pre_transform),
            transform=Helpers.maybe(to_api_obj, self.transform),
        )


class GoldSpec(BaseModel):
    """
    Configuration for gold table in a DataSource.

    Attributes:
        omit_tables (Optional[List[str]]):
            A list of tables to omit from the preset.
        modify_tables (Optional[List[GoldSpec.ModifyTables]]):
            Modifications for existing gold table definitions.
        add_tables (Optional[List[GoldSpec.AddTables]]):
            User defined tables to add to the gold configuration.
    """

    class ModifyTables(BaseModel):
        """
        Modification to gold tables during transformation.

        Attributes:
            name (Optional[str]):
                Table name.
            source_table (Optional[str]):
                Used to match against the preset's gold stanzas input fields.
            custom (Optional[GoldSpec.ModifyTables.Custom]):
                Custom function for modifying tables.
            omit_fields (Optional[List[str]]):
                A list of fields to omit.
            add_fields (Optional[List[FieldSpec]]):
                Fields to add.
            filter (Optional[str]):
                A SQL filter to apply before processing.
            post_filter (Optional[str]):
                A SQL filter to apply after processing.
        """

        class Custom(BaseModel):
            """
            Custom function to use as part of a gold table modification.

            Attributes:
                function (Optional[str]):
                options (Optional[Dict[str, str]]):
            """

            function: Optional[str] = None
            options: Optional[Dict[str, str]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[
                    CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInnerCustom
                ],
            ) -> "GoldSpec.ModifyTables.Custom":
                if obj is None:
                    return None
                return GoldSpec.ModifyTables.Custom(
                    function=obj.function,
                    options=obj.options,
                )

            def to_api_obj(
                self,
            ) -> CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInnerCustom:
                return CoreV1DataSourceSpecSilverTransformPresetOverridesModifyTablesInnerCustom(
                    function=self.function,
                    options=self.options,
                )

        name: Optional[str] = None
        source_table: Optional[str] = None
        custom: Optional["GoldSpec.ModifyTables.Custom"] = None
        omit_fields: Optional[List[str]] = None
        add_fields: Optional[List[FieldSpec]] = None
        filter: Optional[str] = None
        post_filter: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceSpecGoldPresetOverridesModifyTablesInner],
        ) -> "GoldSpec.ModifyTables":
            if obj is None:
                return None
            add_fields = None
            if obj.add_fields is not None:
                add_fields = [FieldSpec.from_api_obj(item) for item in obj.add_fields]
            return GoldSpec.ModifyTables(
                name=obj.name,
                source_table=obj.source_table,
                custom=GoldSpec.ModifyTables.Custom.from_api_obj(obj.custom),
                omit_fields=obj.omit_fields,
                add_fields=add_fields,
                filter=obj.filter,
                post_filter=obj.post_filter,
            )

        def to_api_obj(
            self,
        ) -> CoreV1DataSourceSpecGoldPresetOverridesModifyTablesInner:
            add_fields = None
            if self.add_fields is not None:
                add_fields = [item.to_api_obj() for item in self.add_fields]
            to_api_obj = lambda o: o.to_api_obj()
            return CoreV1DataSourceSpecGoldPresetOverridesModifyTablesInner(
                name=self.name,
                source_table=self.source_table,
                custom=Helpers.maybe(to_api_obj, self.custom),
                omit_fields=self.omit_fields,
                add_fields=add_fields,
                filter=self.filter,
                post_filter=self.post_filter,
            )

    class AddTables(BaseModel):
        """
        Tables to add during gold table transformation.

        Attributes:
            custom (Optional[GoldSpec.AddTables.Custom]):
                Custom function for adding tables.
            name (Optional[str]):
                The name of the table to add.
            source_table (Optional[str]):
                The source table/dataframe for the gold table.
            filter (Optional[str]):
                A SQL filter to apply.
            post_filter (Optional[str]):
                A SQL filter to apply after processing.
            fields (Optional[List[FieldSpec]]):
                Field specifications for the new table.
        """

        class Custom(BaseModel):
            """
            Custom function for adding tables during gold transformation.

            Attributes:
                function (Optional[str]):
                options (Optional[Dict[str, str]]):
            """

            function: Optional[str] = None
            options: Optional[Dict[str, str]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[
                    CoreV1DataSourceSpecGoldPresetOverridesAddTablesInnerCustom
                ],
            ) -> "GoldSpec.AddTables.Custom":
                if obj is None:
                    return None
                return GoldSpec.AddTables.Custom(
                    function=obj.function,
                    options=obj.options,
                )

            def to_api_obj(
                self,
            ) -> CoreV1DataSourceSpecGoldPresetOverridesAddTablesInnerCustom:
                return CoreV1DataSourceSpecGoldPresetOverridesAddTablesInnerCustom(
                    function=self.function,
                    options=self.options,
                )

        name: Optional[str] = None
        source_table: Optional[str] = None
        custom: Optional["GoldSpec.AddTables.Custom"] = None
        filter: Optional[str] = None
        post_filter: Optional[str] = None
        fields: Optional[List[FieldSpec]] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceSpecGoldPresetOverridesAddTablesInner],
        ) -> "GoldSpec.AddTables":
            if obj is None:
                return None
            fields = None
            if obj.fields is not None:
                fields = [FieldSpec.from_api_obj(item) for item in obj.fields]
            return GoldSpec.AddTables(
                custom=GoldSpec.AddTables.Custom.from_api_obj(obj.custom),
                name=obj.name,
                source_table=obj.source_table,
                filter=obj.filter,
                post_filter=obj.post_filter,
                fields=fields,
            )

        def to_api_obj(self) -> CoreV1DataSourceSpecGoldPresetOverridesAddTablesInner:
            fields = None
            if self.fields is not None:
                fields = [item.to_api_obj() for item in self.fields]
            to_api_obj = lambda o: o.to_api_obj()
            return CoreV1DataSourceSpecGoldPresetOverridesAddTablesInner(
                custom=Helpers.maybe(to_api_obj, self.custom),
                name=self.name,
                source_table=self.source_table,
                filter=self.filter,
                post_filter=self.post_filter,
                fields=fields,
            )

    omit_tables: Optional[List[str]] = None
    modify_tables: Optional[List["GoldSpec.ModifyTables"]] = None
    add_tables: Optional[List["GoldSpec.AddTables"]] = None

    @staticmethod
    def from_api_obj(obj: Optional[CoreV1DataSourceSpecGold]) -> "GoldSpec":
        if obj is None:
            return None
        omit_tables = None
        modify_tables = None
        add_tables = None
        if obj.preset_overrides is not None:
            omit_tables = obj.preset_overrides.omit_tables
            if obj.preset_overrides.modify_tables is not None:
                modify_tables = [
                    GoldSpec.ModifyTables.from_api_obj(item)
                    for item in obj.preset_overrides.modify_tables
                ]
            if obj.preset_overrides.add_tables is not None:
                add_tables = [
                    GoldSpec.AddTables.from_api_obj(item)
                    for item in obj.preset_overrides.add_tables
                ]
        return GoldSpec(
            omit_tables=omit_tables,
            modify_tables=modify_tables,
            add_tables=add_tables,
        )

    def to_api_obj(self) -> CoreV1DataSourceSpecGold:
        modify_tables = None
        if self.modify_tables is not None:
            modify_tables = [item.to_api_obj() for item in self.modify_tables]
        add_tables = None
        if self.add_tables is not None:
            add_tables = [item.to_api_obj() for item in self.add_tables]
        return CoreV1DataSourceSpecGold(
            preset_overrides=CoreV1DataSourceSpecGoldPresetOverrides(
                omit_tables=self.omit_tables,
                modify_tables=modify_tables,
                add_tables=add_tables,
            ),
        )


class DataSource(BaseModel):
    """
    A DataSource resource.

    Attributes:
        metadata (Optional[Metadata]):
            Standard object metadata.
        source (Optional[str]):
            The name of the originator of the data.
        source_type (Optional[str]):
            The type of data being imported.
        epoch (Optional[int]):
        schedule (Optional[Schedule]):
            The schedule for data ingestion.
        custom (Optional[DataSource.CustomNotebook]):
            A custom notebook for the datasource.
        primary_key (Optional[PrimaryKey]):
            Primary key configuration of the datasource.
        use_preset (Optional[str]):
            The name of the preset to use for this data source.
        autoloader (Optional[DataSource.Autoloader]):
            Autoloader configuration.
        bronze (Optional[BronzeSpec]):
            Bronze table configuration.
        compute_mode (Optional[str]):
            The compute mode to use for this datasource's job.
        silver (Optional[SilverSpec]):
            Silver transformation configuration.
        gold (Optional[GoldSpec]):
            Gold transformation configuration.
        status (Optional[ResourceStatus]):
            The current status of the datasource.
    """

    class CustomNotebook(BaseModel):
        """
        A custom notebook for generating data.

        Attributes:
            notebook (Optional[str]):
                Path to the notebook in the Databricks workspace.
        """

        notebook: Optional[str] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceSpecCustom],
        ) -> "DataSource.CustomNotebook":
            if obj is None:
                return None
            return DataSource.CustomNotebook(
                notebook=obj.notebook,
            )

        def to_api_obj(self) -> CoreV1DataSourceSpecCustom:
            return CoreV1DataSourceSpecCustom(
                notebook=self.notebook,
            )

    class PrimaryKey(BaseModel):
        """
        PrimaryKey configuration for DataSource

        Attributes:
            time_column (str): column name used as timestamp portion of the sortable synthetic key
            additionalColumns (List[str]): list of columns to compute hashkey over
        """

        time_column: str
        additional_columns: List[str]

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourcePrimaryKeySpec],
        ) -> "DataSource.PrimaryKey":
            if obj is None:
                return None
            return DataSource.PrimaryKey(
                time_column=obj.time_column,
                additional_columns=obj.additional_columns,
            )

        def to_api_obj(self) -> CoreV1DataSourcePrimaryKeySpec:
            return CoreV1DataSourcePrimaryKeySpec(
                timeColumn=self.time_column,
                additionalColumns=self.additional_columns,
            )

    class Autoloader(BaseModel):
        """
        Autoloader configuration for the DataSource.

        Attributes:
            format (Optional[str]):
                The format of the data (e.g., json, parquet, csv, etc.).
            location (str):
                External location for the volume in Unity Catalog.
            schema_file (Optional[str]):
                An optional file containing the schema of the data source.
            cloud_files (Optional[DataSource.Autoloader.CloudFiles]):
                CloudFiles configuration.
        """

        class CloudFiles(BaseModel):
            """
            CloudFiles configuration for the Autoloader.

            Attributes:
                schema_hints_file (Optional[str]):
                schema_hints (Optional[str]):
            """

            schema_hints_file: Optional[str] = None
            schema_hints: Optional[str] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[ContentV1DatasourcePresetAutoloaderCloudFiles],
            ) -> "DataSource.Autoloader.CloudFiles":
                if obj is None:
                    return None
                return DataSource.Autoloader.CloudFiles(
                    schema_hints_file=obj.schema_hints_file,
                    schema_hints=obj.schema_hints,
                )

            def to_api_obj(self) -> ContentV1DatasourcePresetAutoloaderCloudFiles:
                return ContentV1DatasourcePresetAutoloaderCloudFiles(
                    schema_hints_file=self.schema_hints_file,
                    schema_hints=self.schema_hints,
                )

        format: Optional[str] = None
        location: str
        schema_file: Optional[str] = None
        cloud_files: Optional["DataSource.Autoloader.CloudFiles"] = None

        @staticmethod
        def from_api_obj(
            obj: Optional[CoreV1DataSourceAutoloaderSpec],
        ) -> "DataSource.Autoloader":
            if obj is None:
                return None
            return DataSource.Autoloader(
                format=obj.format,
                location=obj.location,
                schema_file=obj.schema_file,
                cloud_files=DataSource.Autoloader.CloudFiles.from_api_obj(
                    obj.cloud_files
                ),
            )

        def to_api_obj(self) -> CoreV1DataSourceAutoloaderSpec:
            return CoreV1DataSourceAutoloaderSpec(
                format=self.format,
                location=self.location,
                schema_file=self.schema_file,
                cloud_files=Helpers.maybe(lambda o: o.to_api_obj(), self.cloud_files),
            )

    metadata: Optional[Metadata] = None
    source: Optional[str] = None
    source_type: Optional[str] = None
    epoch: Optional[int] = None
    schedule: Optional[Schedule] = None
    custom: Optional["DataSource.CustomNotebook"] = None
    primary_key: Optional["DataSource.PrimaryKey"] = None
    use_preset: Optional[str] = None
    use_preset_version: Optional[int] = None
    autoloader: Optional["DataSource.Autoloader"] = None
    compute_mode: Optional[str] = None
    bronze: Optional[BronzeSpec] = None
    silver: Optional[SilverSpec] = None
    gold: Optional[GoldSpec] = None
    status: Optional[ResourceStatus] = None

    @staticmethod
    def from_api_obj(obj: CoreV1DataSource) -> "DataSource":
        return DataSource(
            metadata=Metadata.from_api_obj(obj.metadata),
            source=obj.spec.source,
            source_type=obj.spec.source_type,
            epoch=obj.spec.epoch,
            schedule=Schedule.from_api_obj(obj.spec.schedule),
            custom=DataSource.CustomNotebook.from_api_obj(obj.spec.custom),
            primary_key=DataSource.PrimaryKey.from_api_obj(obj.spec.primary_key),
            use_preset=obj.spec.use_preset,
            use_preset_version=obj.spec.use_preset_version,
            autoloader=DataSource.Autoloader.from_api_obj(obj.spec.autoloader),
            compute_mode=obj.spec.compute_mode,
            bronze=BronzeSpec.from_api_obj(obj.spec.bronze),
            silver=SilverSpec.from_api_obj(obj.spec.silver),
            gold=GoldSpec.from_api_obj(obj.spec.gold),
            status=ResourceStatus.from_api_obj(obj.status),
        )

    def to_api_obj(self) -> CoreV1DataSource:
        to_api_obj = lambda o: o.to_api_obj()
        return CoreV1DataSource(
            api_version="v1",
            kind="DataSource",
            metadata=Helpers.maybe(to_api_obj, self.metadata),
            spec=CoreV1DataSourceSpec(
                source=self.source,
                source_type=self.source_type,
                epoch=self.epoch,
                schedule=Helpers.maybe(to_api_obj, self.schedule),
                custom=Helpers.maybe(to_api_obj, self.custom),
                primary_key=Helpers.maybe(to_api_obj, self.primary_key),
                use_preset=self.use_preset,
                use_preset_version=self.use_preset_version,
                autoloader=Helpers.maybe(to_api_obj, self.autoloader),
                compute_mode=self.compute_mode,
                bronze=Helpers.maybe(to_api_obj, self.bronze),
                silver=Helpers.maybe(to_api_obj, self.silver),
                gold=Helpers.maybe(to_api_obj, self.gold),
            ),
            status=Helpers.maybe(to_api_obj, self.status),
        )



from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import *
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import lit, col as col_, sum as sum_, when
from dasl_client.preset_development.preview_parameters import *
from dasl_client.preset_development.stage import *
from dasl_client.preset_development.errors import *
import yaml
from IPython import get_ipython
from itertools import count


@udf(StringType())
def constant_udf(*args):
    return "<sortable_random_id>"


class PreviewEngine:
    """
    This class deserializes the in-development preset's YAML and performs a series of
    validation steps before attempting to compile each stage's table and execute them
    based on the provided PreviewParameters.

    Upon successful execution, output is generated for each successfully executed
    stage's table operations. Additionally, if Gold stages are computed, their outputs
    are validated against the provided Gold stage tables to ensure compatibility on a
    per-table-name basis with the Unity Catalog.

    For example, a preset Gold stage table named "http_activity" will be checked against
    the corresponding table in the Unity Catalog schema—also named "http_activity" to
    confirm that inserting into the Unity Catalog most likely not cause errors.
    """

    def __init__(
        self, spark: SparkSession, preset_yaml_str: str, ds_params: PreviewParameters
    ):
        """
        Creates the PreviewEngine using the given preset YAML and datasource parameters.
        The YAML is deserialized here and checked to verify whether the requested
        pretransform name, if provided, exists in the preset.

        Instance Attributes:
            ds_params (PreviewParameters): The input datasource's configuration.
            preset (Dict[str, Any]): The deserialized preset YAML.
            pretransform_name (str): The name of the requested pretransform. Defaults to None.
            pre (Stage): Stores the pretransform Stage object internally.
            silver (List[Stage]): Stores the Silver Stage objects internally.
            gold (List[Stage]): Stores the Gold Stage objects internally.
        """
        self._spark = spark
        self._ds_params = ds_params
        self.__stage_exception = {}
        self._preset = yaml.safe_load(preset_yaml_str)
        self._pretransform_name = ds_params._pretransform_name

        self._validate_gold_inputs(
            self._preset.get("silver", None), self._preset.get("gold", None)
        )
        if self._pretransform_name:
            self._validate_pretransform_name(
                self._preset.get("silver", None), self._pretransform_name
            )

        self._pre_bronze = None
        self._bronze = None
        self._pre_silver = None
        self._silver = []
        self._gold = []
        self._result_df_map: Tuple[
            DataFrame, Dict[str, DataFrame], Dict[str, DataFrame]
        ] = (None, {}, {})

    def _validate_pretransform_name(
        self, silver: Dict[str, str], pretransform_name: str
    ) -> None:
        """
        Validates the given pretransform name exists in the provided preset's Silver
        PreTransform stages.
        """
        if not silver:
            raise NoSilverStageProvdedError(", but pretransform name provided")
        if not (silver_pre_transform := silver.get("preTransform", None)):
            raise NoSilverPreTransformStageProvdedError()
        silver_pre_output_names = []
        for table in silver_pre_transform:
            if not (name := table.get("name", None)):
                raise MissingTableFieldError(
                    "Silver pretransform",
                    table.get("name", "<stage missing name>"),
                    "name",
                )
            silver_pre_output_names += [name]
        if pretransform_name not in silver_pre_output_names:
            raise PreTransformNotFound()

    def _validate_gold_inputs(
        self, silver: Dict[str, str], gold: Dict[str, str]
    ) -> None:
        """
        Validate gold tables all have a silver table to input from.
        """
        if not gold:
            return

        if not len(gold):
            return

        if not silver:
            raise NoSilverStageProvdedError(", but gold stage is present")

        gold_input_names = []
        for table in gold:
            if not (input := table.get("input", None)):
                raise MissingTableFieldError(
                    "Gold", table.get("name", "<stage missing name>"), "input"
                )
            gold_input_names += [input]

        if not (silver_transform := silver.get("transform", None)):
            raise NoSilverTransformStageProvdedError()
        silver_output_names = []
        for table in silver_transform:
            if not (name := table.get("name", None)):
                raise MissingTableFieldError(
                    "Silver transform", table.get("name", ""), "name"
                )
            silver_output_names += [name]

        missing_keys = set(gold_input_names) - set(silver_output_names)
        if missing_keys:
            raise MissingSilverKeysError(missing_keys)

    def _compile_stages(self, force_evaluation: bool = False) -> None:
        """
        Creates Stage objects, setting silver pretransform to None if not provided.
        """
        pre_bronze_field_counter = count()
        pre_bronze_name_counter = count()
        pre_bronze_expr_groups = self._preset.get("bronze", {}).get("preTransform", [])
        if pre_bronze_expr_groups:
            tables = [
                {
                    "name": f"Index {next(pre_bronze_name_counter)}",
                    "fields": [
                        {"name": str(next(pre_bronze_field_counter)), "expr": expr}
                        for expr in expr_group
                    ],
                }
                for expr_group in pre_bronze_expr_groups
            ]
            for table in tables:
                self._pre_bronze = [
                    Stage(self._spark, "bronze pretransform", table) for table in tables
                ]

        pretransform = None
        if self._pretransform_name:
            for table in self._preset["silver"]["preTransform"]:
                if table["name"] == self._pretransform_name:
                    self._pre_silver = Stage(self._spark, "silver pretransform", table)
                    break

        self._silver = [
            Stage(
                self._spark,
                "silver transform",
                table,
                force_evaluation=force_evaluation,
            )
            for table in self._preset.get("silver", {}).get("transform", [])
        ]
        self._gold = [
            Stage(self._spark, "gold", table, force_evaluation=force_evaluation)
            for table in self._preset.get("gold", [])
        ]

    def _run(
        self, df: DataFrame, verbose: bool = False
    ) -> Tuple[DataFrame, Dict[str, DataFrame], Dict[str, DataFrame]]:
        """
        Runs all stages, in medallion stage order. This allows prior stage outputs to feed
        into later stage inputs.

        Returns:
            Dataframes containing the output from each run Stage.
        """
        # If we are in silverbronze mode, and an autoloader has been provided, or we are
        # not in silverbronze mode, we need to run the preBronze stage.
        pre_bronze_output = {}
        if (
            self._ds_params._mode != "silverbronze"
            or self._ds_params._autoloader_location
        ):
            if self._pre_bronze:
                for stage in self._pre_bronze:
                    df = stage.run(df)
                    pre_bronze_output[stage._name] = df
        else:
            # We are in silverbronze mode with no autoloader, so we treat first
            # silverbronze table as initial df.
            df = (
                self._spark.table(self._ds_params._bronze_tables[0].get("name", ""))
                .drop("dasl_id")
                .limit(self._ds_params._record_limit)
            )

        if time_col := self._ds_params._time_column:
            df = df.filter(
                f"timestamp({time_col}) >= timestamp('{self._ds_params._start_time}') AND timestamp({time_col}) < timestamp('{self._ds_params._end_time}')"
            )

        df = df.withColumn("dasl_id", constant_udf())

        self._bronze = df

        # Deal with silverbronze table joins.
        # Note: We can blind get here as validation should've caught anything missing.
        if self._ds_params._mode == "silverbronze":
            if alias := self._ds_params._bronze_tables[0].get("alias", None):
                df = df.alias(alias)
            for bronze_table in self._ds_params._bronze_tables[1:]:
                join_df = (
                    spark.table(bronze_table["name"])
                    .drop("dasl_id")
                    .limit(self._ds_params._record_limit)
                )
                if alias := bronze_table.get("alias", None):
                    join_df = join_df.alias(alias)
                df = df.join(
                    join_df,
                    expr(bronze_table["joinExpr"]),
                    bronze_table.get("joinType", "left"),
                )

        if self._pre_silver:
            df = self._pre_silver.run(df)

        silver_output_map = {}
        for table in self._silver:
            silver_output_map[table._name] = table.run(df)

        # Check for silver stage exceptions.
        # NOTE: These exception lists only get populated if force_evaluation is enabled.
        for table in self._silver:
            if exceptions := table.get_exceptions():
                self.__stage_exception[table._name] = exceptions
        if self.__stage_exception:
            raise StageExecutionException("silver", self.__stage_exception, verbose)

        gold_output_map = {}
        for table in self._gold:
            # We store as gold_name/silver_input to prevent clobbering on duplicate gold table use.
            gold_output_map[f"{table._name}/{table._input}"] = table.run(
                silver_output_map[table._input]
            )

        # Check for gold stage exceptions.
        # NOTE: These exception lists only get populated if force_evaluation is enabled.
        for table in self._gold:
            if exceptions := table.get_exceptions():
                self.__stage_exception[table._name] = exceptions
        if self.__stage_exception:
            raise StageExecutionException("gold", self.__stage_exception, verbose)

        return (
            (df, silver_output_map, gold_output_map, pre_bronze_output)
            if self._pre_silver
            else (None, silver_output_map, gold_output_map, pre_bronze_output)
        )

    def __get_sql_type(self, data_type) -> str:
        """
        Helper to convert Spark data type objects to SQL type strings.
        """
        if isinstance(data_type, StringType):
            return "STRING"
        elif isinstance(data_type, IntegerType):
            return "INT"
        elif isinstance(data_type, LongType):
            return "BIGINT"
        elif isinstance(data_type, FloatType):
            return "FLOAT"
        elif isinstance(data_type, DoubleType):
            return "DOUBLE"
        elif isinstance(data_type, BooleanType):
            return "BOOLEAN"
        elif isinstance(data_type, TimestampType):
            return "TIMESTAMP"
        elif isinstance(data_type, DateType):
            return "DATE"
        elif isinstance(data_type, ArrayType):
            return f"ARRAY<{self.__get_sql_type(data_type.elementType)}>"
        elif isinstance(data_type, MapType):
            return f"MAP<{self.__get_sql_type(data_type.keyType)}, {self.__get_sql_type(data_type.valueType)}>"
        elif isinstance(data_type, StructType):
            fields = ", ".join(
                [
                    f"{field.name}: {self.__get_sql_type(field.dataType)}"
                    for field in data_type.fields
                ]
            )
            return f"STRUCT<{fields}>"
        elif isinstance(data_type, VariantType):
            return f"VARIANT"
        else:
            return f"UNKNOWN ({data_type})"

    def __format_gold_column_merge_exception(
        self,
        columns: Dict[str, List[Exception]],
        gold_df: DataFrame,
        verbose: bool = False,
    ):
        """
        Formatter for various exceptions that occur during the merge of gold tables.
        """
        missing_column_flag = False
        for column, info in columns.items():
            # RANT: it is annoying, but basically every exception comes back from the
            # query analyzer as pyspark.errors.exceptions.connect.AnalysisException,
            # so we are forced into this awkward string search.
            str_e = str(info["exception"])
            str_e = str_e.split("JVM")[0] if not verbose else str_e
            if "LEGACY_ERROR_TEMP_DELTA_0007" in str_e:
                print(
                    f"-> Column \"{column}\" of type \"{self.__get_sql_type(info['type'])}\" does not exist in gold table \"{info['table']}\"."
                )
                missing_column_flag = True
            elif "DELTA_FAILED_TO_MERGE_FIELDS" in str_e:
                print(
                    f"-> Column \"{column}\" of type \"{self.__get_sql_type(info['type'])}\" is not compatiable with gold table \"{info['table']}\"'s \"{column}\" of type \"{self.__get_sql_type(gold_df.schema[column].dataType)}\""
                )
            else:
                print(
                    f"-> Column \"{column}\" raised the following unformatted exception when appending to gold table \"{info['table']}\":\n{str_e}"
                )

        if missing_column_flag:
            print(
                f"\nA write to 1 or more non-existent columns occured - available columns are: {', '.join(gold_df.columns)}"
            )

    def _render_output(
        self,
        input_df: DataFrame,
        stage_dataframes: Tuple[
            List[DataFrame], DataFrame, Dict[str, DataFrame], Dict[str, DataFrame]
        ],
        gold_table_catalog: str,
        gold_table_schema: str,
        verbose: bool = False,
    ) -> None:
        """
        Displays formatted HTML output from executed Stages' DataFrames.
        """
        # TODO: Investigate further into using Databricks's style sheets here.

        # Get the Databricks built-in functions out the namespace.
        ipython = get_ipython()
        if ipython is not None:
            displayHTML = ipython.user_ns["displayHTML"]
            display = ipython.user_ns["display"]
        else:
            displayHTML = lambda x: print(x)
            display = lambda x: x.show()

        def d(txt, lvl) -> None:
            displayHTML(
                f"""
                <div style="background-color:
                background-color: rgb(18, 23, 26); padding: 0; margin: 0;">
                    <h{lvl} style="margin: 0; background-color: rgb(244, 234, 229);">{txt}</h{lvl}>
                </div>
                """
            )

        (pre_silver, silver, gold, pre_bronze) = stage_dataframes
        d("Autoloader Input", 1)
        display(input_df)
        d("Bronze Pre-Transform", 1)
        for name, df in pre_bronze.items():
            d(f"{name}", 2)
            display(df)
        d("Silver Pre-Transform", 1)
        if pre_silver:
            display(pre_silver)
        else:
            d("Skipped", 2)
        d("Silver Transform", 1)
        for name, df in silver.items():
            d(f"{name}", 2)
            display(df)
        d("Gold", 1)
        for full_name, df in gold.items():
            d(f"{full_name}", 2)
            d("Stage output", 3)
            display(df)

            # NOTE: Name is stored as Gold_name/Silver_input. So we need to get just the Gold table
            # name that we are comparing the dataframe metadata to.
            name = full_name.split("/")[0]
            fqn_gold_table_name = f"{self.force_apply_backticks(gold_table_catalog)}.{self.force_apply_backticks(gold_table_schema)}.{self.force_apply_backticks(name)}"

            if not self._spark.catalog.tableExists(f"{fqn_gold_table_name}"):
                raise UnknownGoldTableError(name, gold_table_schema)

            # Create a temporary table to perform the type check
            delta_df = self._spark.table(f"{fqn_gold_table_name}").limit(0)
            delta_df.write.mode("overwrite").save(
                f"{self._ds_params.get_autoloader_temp_schema_location()}/{full_name}"
            )

            # Update the params to indicate we've added a testing temp gold table
            self._ds_params.add_gold_schema_table(full_name)

            # Perform the type checks by trying to insert data into the table

            df_columns = df.columns
            df_single_columns = {}
            df_append_exceptions = {}
            for column in df_columns:
                df_single_columns[column] = df.select(column)
            for column, df_single_column in df_single_columns.items():
                try:
                    df_single_column.write.mode("append").save(
                        f"{self._ds_params.get_autoloader_temp_schema_location()}/{full_name}"
                    )
                except Exception as e:
                    df_append_exceptions[column] = {
                        "type": df_single_column.schema[column].dataType,
                        "exception": e,
                        "table": name,
                    }

            self.__format_gold_column_merge_exception(
                df_append_exceptions, delta_df, verbose
            )

            if not df_append_exceptions:
                # alls good. display the output.
                d("Resultant gold table preview", 3)
                unioned_df = delta_df.unionByName(df, allowMissingColumns=True)
                display(unioned_df)

    def is_backtick_escaped(self, name: str) -> bool:
        """
        check if a given (column) name is backtick escaped or not
        :param name: column name
        :return: bool
        """
        return name.startswith("`") and name.endswith("`")

    def force_apply_backticks(self, name: str) -> str:
        """
        forces application of backticks to the given (column) name as a single unit
        if it already has backticks this is a noop
        :param name: column name
        :return: str
        """
        if self.is_backtick_escaped(name):
            return name
        return f"`{name}`"

    def evaluate(
        self,
        gold_table_schema: str,
        display: bool = True,
        force_evaluation: bool = False,
        verbose: bool = False,
    ) -> None:
        """
        Evaluates the loaded preset YAML using the input datasource configuration to load
        records. Finally, checks that the output from the Gold stages is compatible with
        the Unity Catalog Gold tables.
        """
        s = gold_table_schema.split(".")
        if len(s) != 2:
            raise InvalidGoldTableSchemaError(gold_table_schema)
        catalog_name = s[0].lstrip("`").rstrip("`")
        schema_name = s[1].lstrip("`").rstrip("`")
        if any(
            row.catalog == catalog_name
            for row in self._spark.sql("SHOW CATALOGS").collect()
        ):
            if not any(
                row.databaseName == schema_name
                for row in self._spark.sql(
                    f"SHOW SCHEMAS IN `{catalog_name}`"
                ).collect()
            ):
                raise InvalidGoldTableSchemaError(
                    gold_table_schema,
                    f"Schema {schema_name} not found in catalog {catalog_name} or insufficient permissions.",
                )
        else:
            raise InvalidGoldTableSchemaError(
                gold_table_schema,
                f"Catalog {catalog_name} not found or insufficient permissions.",
            )

        # If we are using the autoloader, fetch format from preset and others.
        if self._ds_params._mode == "autoloader" or (
            self._ds_params._mode == "silverbronze"
            and self._ds_params._autoloader_location
        ):
            if self._preset.get("bronze", {}).get("loadAsSingleVariant", False) == True:
                self._ds_params._set_load_as_single_variant()
            if not (autoloader_conf := self._preset.get("autoloader", None)):
                raise MissingAutoloaderConfigError()
            if not (file_format := autoloader_conf.get("format", None)):
                raise AutoloaderMissingFieldError("format")
            self._ds_params._set_autoloader_format(file_format)
            if schemaFile := autoloader_conf.get("schemaFile", None):
                self._ds_params._set_autoloader_schema_file(schemaFile)
            if cloudFiles := autoloader_conf.get("cloudFiles", None):
                if schema_hints := cloudFiles.get("schemaHints", None):
                    self._ds_params._set_autoloader_cloudfiles_schema_hints(
                        schema_hints
                    )
                if schema_hints_file := cloudFiles.get("schemaHintsFile", None):
                    self._ds_params._set_autoloader_cloudfiles_schema_hint_file(
                        schema_hints_file
                    )

        self._compile_stages(force_evaluation=force_evaluation)

        with self._ds_params as df:
            self._result_df_map = self._run(df, verbose)
            if display:
                self._render_output(
                    df,
                    self._result_df_map,
                    self.force_apply_backticks(catalog_name),
                    self.force_apply_backticks(schema_name),
                    verbose,
                )

    def results(
        self,
    ) -> Tuple[DataFrame, DataFrame, Dict[str, DataFrame], Dict[str, DataFrame]]:
        return self._bronze, *self._result_df_map



from importlib.metadata import version


def get_user_agent() -> str:
    """
    A helper function defining the user agent for requests originating from
    the ASL python conn library. We include the version of the API that the
    connection was built off.

    :return: A user-agent string.
    """
    return get_client_identifier()


def get_client_identifier() -> str:
    """
    A helper function defining the client identifier for identifying the python
    client. We include the version of the version of the API that the client
    uses.

    :return: The python client identifier.
    """
    return f"asl-python-client/{version('dasl_api')}"



import json
from collections.abc import Callable
from contextlib import contextmanager

from dasl_api import ApiException
from urllib3.exceptions import MaxRetryError, RequestError


class ConflictError(Exception):
    """
    Simple exception wrapper for 409 errors returned from the API
    """

    def __init__(self, resource: str, identifier: str, message: str) -> None:
        self.resource = resource
        self.identifier = identifier
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Conflict: resource_type='{self.resource}' identifier='{self.identifier}' message='{self.message}'"


class NotFoundError(Exception):
    """
    Simple exception wrapper for 404 errors returned from the API
    """

    def __init__(self, identifier: str, message: str) -> None:
        self.identifier = identifier
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"NotFound: identifier='{self.identifier}' message='{self.message}'"


class BadRequestError(Exception):
    """
    Simple exception wrapper for 400 errors returned from the API
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"BadRequest: message='{self.message}'"


class UnauthorizedError(Exception):
    """
    Simple exception wrapper for 401 errors returned from the API
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Unauthorized: message='{self.message}'"


class ForbiddenError(Exception):
    """
    Simple exception wrapper for 403 errors returned from the API
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Forbidden: message='{self.message}'"


def handle_errors(f: Callable) -> Callable:
    """
    A decorator that handles errors returned from the API.

    :param f: the function that could return an API error
    :return: The output from the callable 'f'. If an Api error was raise,
             re-cast it to a library error before re-raising.
    """

    def error_handler(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ApiException as e:
            body = json.loads(e.body)
            if e.status == 400:
                raise BadRequestError(body["message"])
            if e.status == 401:
                raise BadRequestError(body["message"])
            if e.status == 403:
                raise ForbiddenError(body["message"])
            if e.status == 404:
                raise NotFoundError(body["identifier"], body["message"])
            if e.status == 409:
                raise ConflictError(
                    body["resourceType"], body["identifier"], body["message"]
                )
            else:
                raise e
        except Exception as e:
            raise e

    return error_handler


@contextmanager
def error_handler():
    """
    A context manager that handles errors returned from the API.

    Within the context, if an API error is raised, it is re-cast to a library
    error before re-raising.
    """
    try:
        yield
    except ApiException as e:
        body = json.loads(e.body)
        if e.status == 400:
            raise BadRequestError(body["message"])
        if e.status == 401:
            raise BadRequestError(body["message"])
        if e.status == 403:
            raise ForbiddenError(body["message"])
        if e.status == 404:
            raise NotFoundError(body["identifier"], body["message"])
        if e.status == 409:
            raise ConflictError(
                body["resourceType"], body["identifier"], body["message"]
            )
        else:
            raise e
    except Exception as e:
        raise e



from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import *
from pyspark.sql.dataframe import DataFrame
from typing import Dict, Any, List, Mapping, Tuple
from IPython import get_ipython


class PresetError(Exception):
    pass


class StageExecutionException(PresetError):
    def __init__(
        self,
        medallion_layer="unknown",
        exception_map: Dict[str, List[str]] = {},
        verbose: bool = False,
    ):
        self.exception_map = exception_map
        message = (
            f"Field specification errors encountered in {medallion_layer} stage.\n\n"
        )
        for table, exceptions in exception_map.items():
            message += f"Table: {table}\n"
            count = 1
            for exception in exceptions:
                message += f"Exception {count}:\n{exception.split('JVM')[0] if not verbose else exception}\n\n"
                count += 1
        super().__init__(message)


class InvalidGoldTableSchemaError(PresetError):
    def __init__(self, schema: str, additional_message: str = ""):
        self.schema = schema
        message = (
            f"Malformed gold schema provided {schema}. {additional_message}".strip()
        )
        super().__init__(message)


class NoSilverStageProvdedError(PresetError):
    def __init__(self, additional_msg: str = ""):
        message = f"No silver stage provided{additional_msg}."
        super().__init__(message)


class NoSilverTransformStageProvdedError(PresetError):
    def __init__(
        self,
        message: str = "No silver transform stage provided, but gold stage is present.",
    ):
        super().__init__(message)


class PreTransformNotFound(PresetError):
    def __init__(
        self,
        message: str = "Requested silver pretransform name not found in preset's silver pretransforms.",
    ):
        super().__init__(message)


class NoSilverPreTransformStageProvdedError(PresetError):
    def __init__(
        self,
        message: str = "No silver transform stage provided, but prestransform name provided.",
    ):
        super().__init__(message)


class MissingTableFieldError(PresetError):
    def __init__(self, layer: str, table_name: str, field_name: str):
        self.layer = layer
        self.table_name = table_name
        self.field_name = field_name
        message = f"{layer} stage {table_name} is missing {field_name} field."
        super().__init__(message)


class DuplicateFieldNameError(PresetError):
    def __init__(self, stage: str, stage_name: str, field_name: str):
        self.stage = stage
        self.stage_name = stage_name
        self.field_name = field_name
        message = f"Duplicate field specification name found in {stage} stage {stage_name} named {field_name}."
        super().__init__(message)


class MalformedFieldError(PresetError):
    def __init__(self, stage: str, stage_name: str, field_name: str):
        self.stage = stage
        self.stage_name = stage_name
        self.field_name = field_name
        message = f"Please provide 1 operation only in {stage} stage {stage_name}'s field specification named {field_name}."
        super().__init__(message)


class InvalidLiteralError(PresetError):
    def __init__(self, stage: str, stage_name: str, field_name: str):
        self.stage = stage
        self.stage_name = stage_name
        self.field_name = field_name
        message = f"Literal can only be type string in {stage} stage {stage_name}'s field specification named {field_name}."
        super().__init__(message)


class InvalidFromError(PresetError):
    def __init__(self, stage: str, stage_name: str, field_name: str, reason: str):
        self.stage = stage
        self.stage_name = stage_name
        self.field_name = field_name
        self.reason = reason
        message = f"{reason} in {stage} stage {stage_name}'s field specification named {field_name}."
        super().__init__(message)


class MissingFieldNameError(PresetError):
    def __init__(self, stage: str, stage_name: str):
        self.stage = stage
        self.stage_name = stage_name
        message = (
            f"Field specification in {stage} stage {stage_name} missing name field."
        )
        super().__init__(message)


class MissingSilverKeysError(PresetError):
    def __init__(self, missing_keys: str):
        self.missing_keys = missing_keys
        message = f"Gold table/s have no corresponding input from silver table/s: {missing_keys}"
        super().__init__(message)


class MissingAutoloaderConfigError(PresetError):
    def __init__(
        self,
        message: str = "Autoloader mode selected, but no autoloader configuration found in preset.autoloader.",
    ):
        super().__init__(message)


class AutoloaderMissingFieldError(PresetError):
    def __init__(self, field_name: str):
        self.field_name = field_name
        message = f"Autoloader mode selected, but missing field {field_name} in preset."
        super().__init__(message)


class MissingBronzeTablesError(PresetError):
    def __init__(
        self,
        message: str = "Bronze tables mode selected, but no bronze table definitions provided.",
    ):
        super().__init__(message)


class MissingBronzeTableFieldError(PresetError):
    def __init__(self, field_name: str):
        self.field_name = field_name
        message = f"A bronze table definition is missing a field {field_name} in provided definitions."
        super().__init__(message)


class UnknownGoldTableError(PresetError):
    def __init__(self, table_name: str, schema: str):
        self.table_name = table_name
        self.schema = schema
        message = (
            f"The referenced Gold table name {table_name} does not exist in {schema}."
        )
        super().__init__(message)


class GoldTableCompatibilityError(PresetError):
    def __init__(self, message: str):
        super().__init__(message)


class ReferencedColumnMissingError(PresetError):
    def __init__(self, operation: str, column_name: str):
        self.operation = operation
        self.column_name = column_name
        message = f"The referenced column {column_name} was not found in the dataframe during {operation} operation."
        super().__init__(message)


class MissingJoinFieldError(PresetError):
    def __init__(self, field_name: str):
        self.field_name = field_name
        message = f"Join operation is missing required field {field_name}."
        super().__init__(message)


class MissingUtilityConfigurationFieldError(PresetError):
    def __init__(self, operation: str, field_name: str):
        self.operation = operation
        self.field_name = field_name
        message = f"The required configuration field {field_name} was not suppled in the {operation} operation."
        super().__init__(message)


class DisallowedUtilityConfigurationError(PresetError):
    def __init__(self, operation: str, stage: str):
        self.operation = operation
        self.stage = stage
        message = f"The {operation} utility is disallowed in the {stage} stage."
        super().__init__(message)


class AssertionFailedError(PresetError):
    def __init__(self, expr: str, assertion_message: str, df: DataFrame):
        # Get the Databricks built-in functions out the namespace.
        ipython = get_ipython()
        if ipython is not None:
            display = ipython.user_ns["display"]
        else:
            display = lambda x: x.show()

        self.expr = expr
        self.assertion_message = assertion_message
        self.df = df
        message = f"The above rows failed the assertion expression {expr} with reason: {assertion_message}\n"
        display(df)
        super().__init__(message)



from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import *
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, lit
from dasl_client.preset_development.errors import *

import re

FieldSpec = Dict[str, Any]


class Node:
    """
    Represents a node in a tree structure. This is used to figure out STRUCT types
    and ensure proper subSTRUCT reconciliation at data output.
    """

    pass


class Branch(Node):
    """
    Represents a branch in a tree structure.
    """

    def __init__(self, children: Dict[str, Node] = None):
        self._children: Dict[str, Node] = children if children is not None else {}


class Leaf(Node):
    """
    Represents a leaf node in a tree structure.
    """

    def __init__(self, field: FieldSpec):
        self._field = field


class Stage:
    """
    Stage represents a single stage's table and contains all logic required to correctly perform
    operations as defined in the table's field specifications, utility functions, filters, and more.

    To operate, it takes in a DataFrame and returns a new DataFrame with the defined operations
    applied. The resulting DataFrame can be used in subsequent stages if required.
    """

    __op_list = ["assert", "literal", "from", "alias", "expr", "join"]

    # From fieldSpec valid characters.
    __invalid_char_pattern = re.compile(r"[\s,;{}\(\)\n\t=]")

    def __validate_field_spec(self, fields: List[Dict[str, str]], stage: str):
        names = []
        for field in self._fields:
            # Check for name. If no name check for assert.
            if not (name := field.get("name", None)):
                if not field.get("assert", None):  # Can't walrus em all :/
                    raise MissingFieldNameError(self._stage, self._name)

            # Check this new name does not duplicate an existing.
            if name in names:
                raise DuplicateFieldNameError(self._stage, self._name, name)
            names += [name] if name != None else []

            # Check for only 1 defined operation.
            missing_op_count = [
                spec for spec in [field.get(op, None) for op in self.__op_list]
            ].count(None)
            if (missing_op_count == len(self.__op_list)) or (
                len(self.__op_list) - missing_op_count > 1
            ):
                raise MalformedFieldError(
                    self._stage, self._name, field.get("name", None)
                )

            # Literal must be a string.
            if lit := field.get("literal", None):
                if type(lit) != str:
                    raise InvalidLiteralError(
                        self._stage, self._name, field.get("name", None)
                    )

            # Validate from (makes sure its not an expression, etc.). This mirrors Scala code's validation.
            if frm := field.get("from", None):
                if len(frm) >= 256:
                    raise InvalidFromError(
                        self._stage,
                        self._name,
                        field.get("name", None),
                        "Column name too long",
                    )
                if frm.strip() == "" or self.__invalid_char_pattern.search(frm):
                    raise InvalidFromError(
                        self._stage,
                        self._name,
                        field.get("name", None),
                        "Malformed column name referenced",
                    )

    def __init__(
        self,
        spark: SparkSession,
        stage: str,
        table: Dict[str, any],
        force_evaluation: bool = False,
    ):
        """
        Initializes a Stage object that encapsulates all operations required for a single
        table within a stage.

        Instance Attributes:
            stage (str): The medallion layer stage name.
            name (str): The name of the table.
            filter (str): The filter applied before operations.
            postFilter (str): The filter applied after operations.
            utils (Dict[str, Any]): Utility operations to perform on the DataFrame.
            input (str): The name of the prior stage's table to use as input.
            fields (List[Dict[str, str]]): Field specification operations to apply.
            assertions (List[Dict[str, List[str]]]): Assertions to apply after operations.
        """
        self._spark = spark
        self._stage = stage
        self.__force_evaluation = force_evaluation
        self._name = table.get("name", "")
        self._filter = table.get("filter", "")
        self._postFilter = table.get("postFilter", "")
        self._utils = table.get("utils", {})
        self._input = table.get("input", None)
        self.__exceptions = []

        # The dasl_id does not exist before bronze or when dealing with temp fields.
        fields = (
            [{"name": "dasl_id", "from": "dasl_id"}] + table.get("fields", [])
            if self._stage not in ["temp_fields", "bronze pretransform"]
            else table.get("fields", [])
        )
        self._fields = [
            f for f in fields if (f.get("name", None) and not f.get("assert", None))
        ]
        self._assertions = [f for f in fields if f.get("assert", None)]

        self.__validate_field_spec(self._fields, self._stage)

    def get_exceptions(self) -> List[str]:
        """
        Get the list of exceptions encountered during field spec evaluation.
        """
        return self.__exceptions

    def _referenced_columns(self) -> List[str]:
        """
        Get a list of columns referenced in the table's field specifications.

        Returns:
            A list of referenced columns.
        """
        return [field.get("from") for field in self._fields if field.get("from", None)]

    def _column_names(self) -> List[str]:
        """
        Returns a list of column names referenced in the table's field specifications.

        Returns:
            A list of columns names.
        """
        return [field.get("name") for field in self._fields]

    def _omitted_columns(self) -> List[str]:
        """
        Get omitted columns from the preserve utility function object.

        Returns:
            A list of omitted columns.
        """
        preserve = self._utils.get("unreferencedColumns", None)
        return [] if not preserve else preserve.get("omitColumns", [])

    def _duplicate_prefix(self) -> str:
        """
        Get the prefix to use for duplicate fields. If not provided, returns a default.

        Returns:
            The duplicate prefix string.
        """
        preserve = self._utils.get("unreferencedColumns", None)
        return "d_" if not preserve else preserve.get("duplicatePrefix", "d_")

    def json_extract_boilerplate(
        self, df: DataFrame, source: str, omit_fields: List[str], duplicate_prefix: str
    ) -> Tuple[StructType, List[str]]:
        """
        Processes the common schema and column information needed from a target JSON
        containing column.

        Returns:
            The schema and columns extracted from the target JSON.
        """
        target_col = source
        existing_columns = df.columns
        if target_col not in existing_columns:
            raise ReferencedColumnMissingError("jsonExtract", target_col)
        schema = self._spark.sql(
            f"SELECT schema_of_json_agg({self.auto_backtick(target_col)}) AS sc FROM {{df}}",
            df=df,
        ).collect()[0][0]
        extract_df = self._spark.createDataFrame(data=[], schema=schema)
        columns = extract_df.columns
        columns = [
            self.auto_backtick(f"extract.{col}") + f" AS {self.auto_backtick(col)}"
            for col in columns
            if col not in omit_fields and col not in existing_columns
        ]
        columns += [
            self.auto_backtick(f"extract.{col}")
            + f" AS {self.auto_backtick(duplicate_prefix + col)}"
            for col in columns
            if col not in omit_fields and col in existing_columns
        ]
        return (schema, columns)

    def json_extract(
        self, df: DataFrame, schema: str, columns: List[str], target_col: str
    ) -> DataFrame:
        """
        Prefroms JSON extraction to new fields in the DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        return (
            df.selectExpr(
                "*",
                f"from_json({self.auto_backtick(target_col)}, '{schema}') AS extract",
            )
            .selectExpr("*", *columns)
            .drop("extract")
        )

    def json_extract_embed_column(
        self,
        df: DataFrame,
        schema: str,
        omit_fields: List[str],
        target_col: str,
        name: str,
    ) -> DataFrame:
        """
        Performs JSON extraction embedding new JSON fields to a single new field in
        the DataFrame. (Serialized as JSON.)

        Returns:
            A DataFrame with the resultant operation's records.
        """
        extract_df = self._spark.createDataFrame(data=[], schema=schema)
        schema = extract_df.drop(omit_fields).schema.simpleString()
        return df.selectExpr(
            "*",
            f"from_json({self.auto_backtick(target_col)}, '{schema}') AS {self.auto_backtick(name)}",
        )

    def preserved_columns(
        self, df: DataFrame
    ) -> Tuple[List[str], List[str], List[str]]:
        """
        Performs unreferenced field preservation to new fields in the DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        # We do not want to preserve temporary fields.
        temp_columns = []
        if temp_fields := self._utils.get("temporaryFields", None):
            temp_columns = [f.get("name") for f in temp_fields]

        referenced_columns = self._referenced_columns()
        omitted_columns = self._omitted_columns()
        # omitted and referenced cols MAY be wrapped in backticks
        # so we need to check for their existence here on top of just plain col name
        preserved_columns = [
            col
            for col in df.columns
            if col not in referenced_columns
            and col not in omitted_columns
            and self.force_apply_backticks(col) not in referenced_columns
            and self.force_apply_backticks(col) not in omitted_columns
        ]
        duplicate_prefix = self._duplicate_prefix()
        column_names = self._column_names()
        duplicate_renames = [
            f"{self.auto_backtick(col)} AS {self.auto_backtick(duplicate_prefix + col)}"
            for col in preserved_columns
            if col in column_names
        ]
        preserved_columns = [
            col
            for col in preserved_columns
            if (col not in column_names and col not in temp_columns)
        ]
        column_names = [col for col in column_names if col not in preserved_columns]

        return (preserved_columns, duplicate_renames, column_names)

    def preserved_columns_embed_column(self, df) -> List[str]:
        """
        Performs unreferenced field preservation to new a single field in the DataFrame.
        (Serialized as JSON.)

        Returns:
            A DataFrame with the resultant operation's records.
        """
        referenced_columns = self._referenced_columns()
        omitted_columns = self._omitted_columns()
        preserved_columns = [
            col
            for col in df.columns
            if col not in referenced_columns and col not in omitted_columns
        ]
        return preserved_columns

    def insert_path(
        self, tree: Dict[str, Node], path: List[str], field: FieldSpec
    ) -> None:
        """
        Inserts a field specification into a tree of nodes.
        """
        if not path:
            return

        head = path[0]
        if len(path) == 1:
            tree[head] = Leaf(field)
        else:
            if head in tree and isinstance(tree[head], Branch):
                sub_tree = tree[head]._children
            else:
                sub_tree = {}
            self.insert_path(sub_tree, path[1:], field)
            tree[head] = Branch(sub_tree)

    def parse_to_tree(self, fields: List[FieldSpec]) -> Dict[str, Node]:
        """
        Parses a list of field specifications into a tree of nodes.
        """
        tree: Dict[str, Node] = {}
        for field in fields:
            name = field.get("name")
            if name is None:
                continue
            path = name.split(".")
            self.insert_path(tree, path, field)
        return tree

    def cast_to_expr(self, field: FieldSpec, name: str) -> str:
        """
        Casts a field specification into a SELECT expression.

        Returns:
            The SQL expression for the field.
        """
        if field.get("from", None):
            # check that the from column exists in the df?
            return f"{self.auto_backtick(field['from'])} AS {self.auto_backtick(name)}"
        elif field.get("literal", None):
            return f"{repr(field['literal'])} AS {self.auto_backtick(name)}"
        elif field.get("expr", None) is not None:
            expr = field["expr"].replace("\\", "\\\\")
            # If we are in a bronze pretransform, we do not want the fieldSpec.name.
            return (
                f"{expr} AS {self.auto_backtick(name)}"
                if self._stage != "bronze pretransform"
                else expr
            )
        else:
            return ""

    def is_backtick_escaped(self, name: str) -> bool:
        """
        check if a given (column) name is backtick escaped or not
        :param name: column name
        :return: bool
        """
        return name.startswith("`") and name.endswith("`")

    def auto_backtick(self, name: str) -> str:
        """
        auto-backtick given name in case it isn't already backtick escaped.
        if the name contains dots it will get split and each component backticked individually.
        Returns the name wrapped in backticks or the passed name if it already had backticks.
        :param name: column name
        :return: str
        """
        if self.is_backtick_escaped(name):
            return name
        parts = name.split(".")
        return ".".join(list(map(lambda s: f"`{s}`", parts)))

    def force_apply_backticks(self, name: str) -> str:
        """
        forces application of backticks to the given (column) name as a single unit
        if it already has backticks this is a noop
        :param name: column name
        :return: str
        """
        if self.is_backtick_escaped(name):
            return name
        return f"`{name}`"

    def process_node(self, name: str, node: Node) -> str:
        """
        Processes a single node in a tree of nodes.

        Returns:
            The STRUCT or SELECT SQL expression for the node.
        """
        if isinstance(node, Leaf):
            return self.cast_to_expr(node._field, name)
        elif isinstance(node, Branch):
            fields_list = []
            for child_name, child_node in node._children.items():
                child_expr = self.process_node(child_name, child_node)
                fields_list.append(f"{child_expr}")
            joined_fields = ",\n".join(fields_list)
            return f"struct(\n{joined_fields}\n) AS {self.auto_backtick(name)}"
        else:
            return ""

    def parse_to_string(self, nested_tree: Dict[str, Node]) -> str:
        """
        Processes the nested tree representation to a valid SELECT expression.

        Returns:
            The SQL expression.
        """
        lines = []
        for name, node in nested_tree.items():
            processed = self.process_node(name, node)
            wrapped = f"\n{processed}\n"
            lines.append(wrapped)
        return lines

    def render_fields(self, fields: List[FieldSpec]) -> str:
        """
        Renders a list of field specifications containing both simple and
        STRUCT references into valid, STRUCT cognicient, SELECT expressions.
        if a nested field is wrapped in backticks it will be treated as a simple field
        for example field of name `col.with.dots` will NOT be treated as nested field.

        Returns:
            The SQL expression.
        """
        simple_fields = [
            f
            for f in fields
            if "." not in f["name"] or self.is_backtick_escaped(f["name"])
        ]
        nested_fields = [
            f
            for f in fields
            if "." in f["name"] and not self.is_backtick_escaped(f["name"])
        ]

        result_parts = []
        for field in simple_fields:
            expr_str = self.cast_to_expr(field, field["name"])
            result_parts.append(f"{expr_str}")

        if nested_fields:
            tree = self.parse_to_tree(nested_fields)
            nested_str = self.parse_to_string(tree)
            result_parts.append(nested_str)

        return [p for p in result_parts if p is not None and len(p) > 0]

    def select_expr(self, df: DataFrame) -> str:
        """
        Renders all field specification operations that result in a SELECT expression
        after filtering, but before post-filtering and aliasing.

        Returns:
            The SQL expression.
        """
        select_fields = self.render_fields(self._fields)

        if preserve := self._utils.get("unreferencedColumns", None):
            if self._stage == "gold":  # No utils run in gold.
                raise DisallowedUtilityConfigurationError("unreferencedColumns", "gold")
            should_preserve = preserve.get("preserve", None)
            if type(should_preserve) != bool:
                raise MissingUtilityConfigurationFieldError(
                    "unreferencedColumns", "preserve"
                )
            if should_preserve:
                if embed_col := preserve.get("embedColumn", None):
                    preserved_columns = self.preserved_columns_embed_column(df)
                    # preserved_columns is obtained from df.columns
                    # applying backticks to all of them is OK here
                    # since they will never use "obj.key" to reference nested fields of structs
                    # so we just go ahead and apply backticks to all across the board
                    colType = preserve.get("embedColumnType", "struct")
                    if colType == "struct":
                        select_fields += [
                            f"struct({', '.join(list(map(lambda x: self.force_apply_backticks(x), preserved_columns)))}) AS {self.auto_backtick(embed_col)}"
                        ]
                    elif colType == "json":
                        select_fields += [
                            f"to_json(struct({', '.join(list(map(lambda x: self.force_apply_backticks(x), preserved_columns)))})) AS {self.auto_backtick(embed_col)}"
                        ]
                    elif colType == "variant":
                        select_fields += [
                            f"parse_json(to_json(struct({', '.join(list(map(lambda x: self.force_apply_backticks(x), preserved_columns)))}))) AS {self.auto_backtick(embed_col)}"
                        ]
                    else:
                        raise UnknownUtilityConfigurationFieldError(
                            "embedColumnType", "unreferencedColumns"
                        )
                else:
                    (
                        preserved_columns,
                        duplicate_renames,
                        column_names,
                    ) = self.preserved_columns(df)
                    # see note above: same here - apply backticks to all columns across the board
                    select_fields += list(
                        map(lambda x: self.force_apply_backticks(x), preserved_columns)
                    )
                    select_fields += list(
                        map(lambda x: self.force_apply_backticks(x), duplicate_renames)
                    )

        return ["*"] + select_fields if self._stage == "temp_fields" else select_fields

    def run_filter(self, df: DataFrame) -> DataFrame:
        """
        Runs filter operations on the provided DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        if self._filter:
            df = df.filter(self._filter)
        return df

    def run_json_extract(self, df: DataFrame) -> DataFrame:
        """
        Runs JSON extract utility operations on the provided DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        if json_extracts := self._utils.get("jsonExtract", None):
            if self._stage == "gold":  # No utils run in gold.
                raise DisallowedUtilityConfigurationError("jsonExtract", "gold")
            for json_extract in json_extracts:
                source = json_extract.get("source")
                if not source:
                    raise MissingUtilityConfigurationFieldError("jsonExtract", "source")
                omit_fields = json_extract.get("omitFields", [])
                duplicate_prefix = json_extract.get("duplicatePrefix", "d_")
                schema, columns = self.json_extract_boilerplate(
                    df, source, omit_fields, duplicate_prefix
                )
                if name := json_extract.get("embedColumn", None):
                    df = self.json_extract_embed_column(
                        df, schema, omit_fields, source, name
                    )
                else:
                    df = self.json_extract(df, schema, columns, source)
        return df

    def run_select_expr(self, df: DataFrame, select_fields: List[str]) -> DataFrame:
        """
        Runs select operations (preserver, from, literal, expr) on the provided DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        # Join columns, processed before this, need to be included in the
        # dataframe too. So we append their output names to the fields
        # selected.
        joins_cols = []
        for field in self._fields:
            if field.get("join", None):
                joins_cols += [field["name"]]

        expressions = select_fields + joins_cols
        if self.__force_evaluation:
            for expression in expressions:
                try:
                    df.selectExpr(expression).collect()  # Collect to make unlazy.
                except Exception as e:
                    self.__exceptions += [f"expression: {expression}\n{str(e)}"]
        return df.selectExpr(expressions)

    def run_joins(self, df: DataFrame) -> DataFrame:
        """
        Runs joins operations on the provided DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        joins = []
        for field in self._fields:
            if field.get("join", None):
                joins += [field]

        for field in joins:
            join = field.get("join")
            lhs = join.get("lhs")
            if not lhs:
                raise MissingJoinFieldError("lhs")
            rhs = join.get("rhs")
            if not rhs:
                raise MissingJoinFieldError("rhs")
            select = join.get("select")
            if not select:
                raise MissingJoinFieldError("select")

            if table := join.get("withTable", None):
                df_joined = self._spark.table(table)
                df = (
                    df.alias("tmp")
                    .join(df_joined, on=[df[lhs] == df_joined[rhs]], how="left")
                    .selectExpr(
                        "tmp.*", f"{select} AS {self.auto_backtick(field.get('name'))}"
                    )
                )
            elif csv := join.get("withCSV", None):
                if path := csv.get("path", None):
                    df_joined = self._spark.read.csv(
                        path, header=True, inferSchema=True
                    )
                    df = (
                        df.alias("tmp")
                        .join(df_joined, on=[df[lhs] == df_joined[rhs]], how="left")
                        .selectExpr(
                            "tmp.*",
                            f"{select} AS {self.auto_backtick(field.get('name'))}",
                        )
                    )
            else:
                raise MissingJoinFieldError("withTable or withCSV (please supply 1)")
        return df

    def run_aliases(self, df: DataFrame) -> DataFrame:
        """
        Runs alias operations on the provided DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        for field in self._fields:
            if field.get("alias", None):
                df = df.selectExpr(
                    "*",
                    f"{self.auto_backtick(field.get('alias'))} AS {self.auto_backtick(field.get('name'))}",
                )
        return df

    def run_assertions(self, df: DataFrame) -> DataFrame:
        """
        Runs assert operations on the provided DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        for assertions in self._assertions:
            for assertion in assertions.get("assert"):
                failing_rows = df.filter(f"NOT ({assertion['expr']})")
                if not failing_rows.isEmpty():
                    raise AssertionFailedError(
                        assertion["expr"], assertion.get("message", ""), failing_rows
                    )
        return df

    def run_post_filter(self, df: DataFrame) -> DataFrame:
        """
        Runs postFilter operations on the provided DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        if self._postFilter:
            df = df.filter(self._postFilter)
        return df

    def run_temp_fields(self, df: DataFrame) -> DataFrame:
        """
        Runs temporary field expressions on the provided DataFrame.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        if temp_fields := self._utils.get("temporaryFields", None):
            if self._stage == "gold":  # No utils run in gold.
                raise DisallowedUtilityConfigurationError("temporaryFields", "gold")
            df = Stage(self._spark, "temp_fields", {"fields": temp_fields}).run(df)
        return df

    def run(self, df: DataFrame) -> DataFrame:
        """
        Runs all provided preset operations in the provided stage's table.

        Returns:
            A DataFrame with the resultant operation's records.
        """
        df = self.run_filter(df)
        df = self.run_temp_fields(df)
        df = self.run_json_extract(df)
        df = self.run_joins(df)
        df = self.run_select_expr(df, self.select_expr(df))
        df = self.run_aliases(df)
        df = self.run_assertions(df)
        df = self.run_post_filter(df)
        return df



from pyspark.sql import DataFrame
from typing import List, Optional

from dasl_api import CoreV1RenderedRuleTables


class ExecRule:
    """
    ExecRule result object allowing access to and clean up of tables
    created as part of the rule rendering endpoint. While the table
    names are exposed as attributes, there are helper functions for
    fetching the contents of the most common tables (i.e. notables
    and observables). In general, it won't be necessary to access
    these attributes. Note that you must call the cleanup function
    when you are done with an instance of this class or else tables
    created as part of rendering and running the rule will be left
    orphaned in your workspace.

    Attributes:
        notables_table (str):
            name of table where notables for the rule execution
            can be found.
        observables_table (str):
            name of table where observables for the rule execution
            can be found.
        opals_table (str):
            name of table where operational alerts for the rule execution
            can be found.
        data_metrics_table (str):
            name of table where rule metrics for the rule execution
            can be found.
        stream_metrics_table (str):
            name of table where stream metrics for the rule execution
            can be found.
        observables_acc_table (str):
            name of table where observables aggregation for the rule
            execution can be found.
    """

    def __init__(self, spark, tables: CoreV1RenderedRuleTables):
        self.spark = spark
        self.notables_table = tables.notables
        self.observables_table = tables.observables
        self.opals_table = tables.operational_alerts
        self.data_metrics_table = tables.data_metrics
        self.stream_metrics_table = tables.stream_metrics
        self.observables_acc_table = tables.observables_accumulation

    def _all_tables(self) -> List[str]:
        return [
            self.notables_table,
            self.observables_table,
            self.opals_table,
            self.data_metrics_table,
            self.stream_metrics_table,
            self.observables_acc_table,
        ]

    def cleanup(self):
        """
        Clean up when done with this ExecRule instance. This method
        cleans up temporarily allocated tables used to store the
        results of the rule execution. Unless you need to preserve
        the results for some reason, you must call this method or
        the temporary tables will be orphaned in your workspace.
        """
        for table in self._all_tables():
            self.spark.sql(f"DROP TABLE IF EXISTS {table}")

    def notables(self, limit: Optional[int] = None) -> DataFrame:
        """
        Return the contents of the notables table.

        :param limit: optional limit to the number of rows returned.
        :returns: DataFrame containing the notables table rows.
        """
        query = f"SELECT * FROM {self.notables_table}"
        if limit is not None:
            query = f"{query} LIMIT {limit}"
        return self.spark.sql(query)

    def observables(self, limit: Optional[int] = None) -> DataFrame:
        """
        Return the contents of the observables table.

        :param limit: optional limit to the number of rows returned.
        :returns: DataFrame containing the observables table rows.
        """
        query = f"SELECT * FROM {self.observables_table}"
        if limit is not None:
            query = f"{query} LIMIT {limit}"
        return self.spark.sql(query)



from pydantic import BaseModel
from typing import Dict, Iterable, List, Optional, Union
from datetime import datetime, timezone
import yaml

from dasl_api import (
    CoreV1Rule,
    CoreV1RuleSpec,
    CoreV1RuleSpecMetadata,
    CoreV1RuleSpecMetadataMitreInner,
    CoreV1RuleSpecMetadataResponse,
    CoreV1RuleSpecMetadataResponsePlaybooksInner,
    CoreV1RuleSpecInput,
    CoreV1RuleSpecInputBatch,
    CoreV1RuleSpecInputStream,
    CoreV1RuleSpecInputStreamCustom,
    CoreV1RuleSpecInputStreamTablesInner,
    CoreV1RuleSpecInputStreamTablesInnerWatermark,
    CoreV1RuleObservable,
    CoreV1RuleObservableRisk,
    CoreV1RuleSpecOutput,
    CoreV1RuleSpecCollate,
    CoreV1RuleSpecInputBatchCustom,
)

from .helpers import Helpers
from .types import Metadata, ResourceStatus, Schedule


class Rule(BaseModel):
    """
    Rules define how to generate notables from input data.

    Attributes:
        metadata (Optional[Metadata]):
            Standard object metadata.
        rule_metadata (Optional[Rule.RuleMetadata]):
            The rule configuration metadata.
        schedule (Schedule):
            The rule schedule.
        compute_mode (Optional[str]):
            The compute mode to use for this rule's job.
        input (Rule.Input):
            The rule input configuration.
        observables (Optional[List[Rule.Observable]]):
            A list of observables.
        output (Rule.Output):
            The rule output configuration.
        collate (Optional[Rule.Collate]):
            The collate configuration.
        status (Optional[ResourceStatus]):
            The current status of the rule.
    """

    class RuleMetadata(BaseModel):
        """
        RuleMetadata object wrapping CoreV1RuleSpecMetadata.

        Attributes:
            version (Optional[Union[float, int]]):
                The current version of the rule.
            category (Optional[str]):
                The category of this detection. The available values are
                configured in workspace config.
            severity (Optional[str]):
                The threat level associated with the notable.
            fidelity (Optional[str]):
                Fidelity is used to capture the maturity of a rule. Newly
                created, untested rules should be marked as Investigative
                fidelity. Older, more well-tested rules should be marked as
                High fidelity. This helps an analyst determine how likely a
                notable is a false positive.
            mitre (Optional[List[Rule.RuleMetadata.Mitre]]):
                Mitre ATT&CK tactic information.
            objective (Optional[str]):
                A longer form description of what this rule is attempting to
                detect (objectMeta.comment is a summary).
            response (Optional[Rule.RuleMetadata.Response]):
                Response configuration for the rule.
        """

        class Mitre(BaseModel):
            """
            Mitre ATT&CK details associated with a Rule.

            Attributes:
                taxonomy (Optional[str]):
                    Mitre ATT&CK taxonomy.
                tactic (Optional[str]):
                    Mitre ATT&CK tactic.
                technique_id (Optional[str]):
                    The Mitre ATT&CK technique identifier.
                technique (Optional[str]):
                    The Mitre ATT&CK technique human-readable name.
                sub_technique_id (Optional[str]):
                    The Mitre ATT&CK sub-technique identifier.
                sub_technique (Optional[str]):
                    The Mitre ATT&CK sub-technique human-readable name.
            """

            taxonomy: Optional[str] = None
            tactic: Optional[str] = None
            technique_id: Optional[str] = None
            technique: Optional[str] = None
            sub_technique_id: Optional[str] = None
            sub_technique: Optional[str] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1RuleSpecMetadataMitreInner],
            ) -> "Rule.RuleMetadata.Mitre":
                if obj is None:
                    return None
                return Rule.RuleMetadata.Mitre(
                    taxonomy=obj.taxonomy,
                    tactic=obj.tactic,
                    technique_id=obj.technique_id,
                    technique=obj.technique,
                    sub_technique_id=obj.sub_technique_id,
                    sub_technique=obj.sub_technique,
                )

            def to_api_obj(self) -> CoreV1RuleSpecMetadataMitreInner:
                return CoreV1RuleSpecMetadataMitreInner(
                    taxonomy=self.taxonomy,
                    tactic=self.tactic,
                    technique_id=self.technique_id,
                    technique=self.technique,
                    sub_technique_id=self.sub_technique_id,
                    sub_technique=self.sub_technique,
                )

        class Response(BaseModel):
            """
            Details regarding how to respond to a notable generated
            by the Rule.

            Attributes:
                guidelines (Optional[str]):
                    Response guidelines.
                playbooks (Optional[List[Rule.RuleMetadata.Response.Playbook]]):
                    Suggested response playbooks.
            """

            class Playbook(BaseModel):
                """
                Databricks notebook and template values to generate a playbook
                for the analyst from the notable.

                Attributes:
                    notebook (Optional[str]):
                        Notebook to run.
                    options (Optional[Dict[str, str]]):
                        These are templated, if they contain ${} this will be
                        filled in using the notable.
                """

                notebook: Optional[str] = None
                options: Optional[Dict[str, str]] = None

                @staticmethod
                def from_api_obj(
                    obj: Optional[CoreV1RuleSpecMetadataResponsePlaybooksInner],
                ) -> "Rule.RuleMetadata.Response.Playbook":
                    if obj is None:
                        return None
                    return Rule.RuleMetadata.Response.Playbook(
                        notebook=obj.notebook,
                        options=obj.options,
                    )

                def to_api_obj(self) -> CoreV1RuleSpecMetadataResponsePlaybooksInner:
                    return CoreV1RuleSpecMetadataResponsePlaybooksInner(
                        notebook=self.notebook,
                        options=self.options,
                    )

            guidelines: Optional[str] = None
            playbooks: Optional[List["Rule.RuleMetadata.Response.Playbook"]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1RuleSpecMetadataResponse],
            ) -> "Rule.RuleMetadata.Response":
                if obj is None:
                    return None
                playbooks = None
                if obj.playbooks is not None:
                    playbooks = [
                        Rule.RuleMetadata.Response.Playbook.from_api_obj(item)
                        for item in obj.playbooks
                    ]
                return Rule.RuleMetadata.Response(
                    guidelines=obj.guidelines,
                    playbooks=playbooks,
                )

            def to_api_obj(self) -> CoreV1RuleSpecMetadataResponse:
                playbooks = None
                if self.playbooks is not None:
                    playbooks = [item.to_api_obj() for item in self.playbooks]
                return CoreV1RuleSpecMetadataResponse(
                    guidelines=self.guidelines,
                    playbooks=playbooks,
                )

        version: Optional[Union[float, int]] = None
        category: Optional[str] = None
        severity: Optional[str] = None
        fidelity: Optional[str] = None
        mitre: Optional[List["Rule.RuleMetadata.Mitre"]] = None
        objective: Optional[str] = None
        response: Optional["Rule.RuleMetadata.Response"] = None

        @staticmethod
        def from_api_obj(obj: Optional[CoreV1RuleSpecMetadata]) -> "Rule.RuleMetadata":
            if obj is None:
                return None
            mitre = None
            if obj.mitre is not None:
                mitre = [
                    Rule.RuleMetadata.Mitre.from_api_obj(item) for item in obj.mitre
                ]
            return Rule.RuleMetadata(
                version=obj.version,
                category=obj.category,
                severity=obj.severity,
                fidelity=obj.fidelity,
                mitre=mitre,
                objective=obj.objective,
                response=Rule.RuleMetadata.Response.from_api_obj(obj.response),
            )

        def to_api_obj(self) -> CoreV1RuleSpecMetadata:
            mitre = None
            if self.mitre is not None:
                mitre = [item.to_api_obj() for item in self.mitre]
            return CoreV1RuleSpecMetadata(
                version=self.version,
                category=self.category,
                severity=self.severity,
                fidelity=self.fidelity,
                mitre=mitre,
                objective=self.objective,
                response=Helpers.maybe(lambda o: o.to_api_obj(), self.response),
            )

    class Input(BaseModel):
        """
        Specification of input data for the Rule.

        Attributes:
            stream (Optional[Rule.Input.Stream]):
                Input if the Rule should operate on streaming input data.
            batch (Optional[Rule.Input.Batch]):
                Input if the rule should operate on batched input data.
        """

        class CustomStream(BaseModel):
            """
            Specification of a stream custom notebook for generating input to
            the Rule.

            Attributes:
                notebook (Optional[str]):
                options (Optional[Dict[str, str]]):
            """

            notebook: Optional[str] = None
            options: Optional[Dict[str, str]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1RuleSpecInputStreamCustom],
            ) -> "Rule.Input.CustomStream":
                if obj is None:
                    return None
                return Rule.Input.CustomStream(
                    notebook=obj.notebook,
                    options=obj.options,
                )

            def to_api_obj(self) -> CoreV1RuleSpecInputStreamCustom:
                return CoreV1RuleSpecInputStreamCustom(
                    notebook=self.notebook,
                    options=self.options,
                )

        class CustomBatch(BaseModel):
            """
            Specification of a batch custom notebook for generating input to
            the Rule.

            Attributes:
                notebook (Optional[str]):
                options (Optional[Dict[str, str]]):
            """

            notebook: Optional[str] = None
            options: Optional[Dict[str, str]] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1RuleSpecInputBatchCustom],
            ) -> "Rule.Input.CustomBatch":
                if obj is None:
                    return None
                return Rule.Input.CustomBatch(
                    notebook=obj.notebook,
                    options=obj.options,
                )

            def to_api_obj(self) -> CoreV1RuleSpecInputBatchCustom:
                return CoreV1RuleSpecInputBatchCustom(
                    notebook=self.notebook,
                    options=self.options,
                )

        class Stream(BaseModel):
            """
            Specification for streaming input from a table or tables,
            from a custom notebook, or directly from a SQL statement.

            Attributes:
                tables (Optional[List[Rule.Input.Stream.Table]]):
                    List of input tables and join rules.
                filter (Optional[str]):
                    A filter expression to be applied to the input stream.
                    Note that this cannot be used in conjunction with a
                    custom SQL expression (i.e. the sql field).
                sql (Optional[str]):
                    A custom SQL expression to apply to the input stream
                    before matching. Note that this cannot be used in
                    conjunction with a filter expression (i.e. the
                    filter member).
                custom (Optional[Rule.Input.CustomStream]):
                starting_timestamp (Optional[datetime]):
                    Starting timestamp for streaming input data. If this
                    value is not specified, then the timestamp when this rule
                    was created will be used. This setting is used to determine
                    the starting point for streaming historical data, and only
                    applies on the first run of the rule. Once some data has
                    been streamed and a checkpoint has been created, this
                    setting no longer has any impact.
            """

            class Table(BaseModel):
                """
                Specification of a table for streaming input.

                Attributes:
                    name (Optional[str]):
                        Name of the table.
                    watermark (Optional[Rule.Input.Stream.Table.Watermark]):
                        Watermark configuration.
                    alias (Optional[str]):
                        Alias name for the table.
                    join_type (Optional[str]):
                        For tables other than the first, how to join to the
                        preceding table.
                    join_expr (Optional[str]):
                        For tables other than the first, the join condition
                        expression to join with the preceding table.
                    streaming (Optional[bool]):
                        For tables other than the first, is this a streaming
                        join or static. Default is false, except on the first
                        table.
                """

                class Watermark(BaseModel):
                    """
                    Watermark for a streaming input table.

                    Attributes:
                        event_time_column (str):
                            Which column is the event time for the delay
                            threshold.
                        delay_threshold (str):
                            A time duration string for the watermark delay.
                        drop_duplicates (Optional[List[str]]):
                            Pass into pyspark dropDuplicates (effectively
                            columns for group by).
                    """

                    event_time_column: str
                    delay_threshold: str
                    drop_duplicates: Optional[List[str]] = None

                    @staticmethod
                    def from_api_obj(
                        obj: Optional[CoreV1RuleSpecInputStreamTablesInnerWatermark],
                    ) -> "Rule.Input.Stream.Table.Watermark":
                        if obj is None:
                            return None
                        return Rule.Input.Stream.Table.Watermark(
                            event_time_column=obj.event_time_column,
                            delay_threshold=obj.delay_threshold,
                            drop_duplicates=obj.drop_duplicates,
                        )

                    def to_api_obj(
                        self,
                    ) -> CoreV1RuleSpecInputStreamTablesInnerWatermark:
                        return CoreV1RuleSpecInputStreamTablesInnerWatermark(
                            event_time_column=self.event_time_column,
                            delay_threshold=self.delay_threshold,
                            drop_duplicates=self.drop_duplicates,
                        )

                name: Optional[str] = None
                watermark: Optional["Rule.Input.Stream.Table.Watermark"] = None
                alias: Optional[str] = None
                join_type: Optional[str] = None
                join_expr: Optional[str] = None
                streaming: Optional[bool] = None

                @staticmethod
                def from_api_obj(
                    obj: Optional[CoreV1RuleSpecInputStreamTablesInner],
                ) -> "Rule.Input.Stream.Table":
                    if obj is None:
                        return None
                    return Rule.Input.Stream.Table(
                        name=obj.name,
                        watermark=Rule.Input.Stream.Table.Watermark.from_api_obj(
                            obj.watermark
                        ),
                        alias=obj.alias,
                        join_type=obj.join_type,
                        join_expr=obj.join_expr,
                        streaming=obj.streaming,
                    )

                def to_api_obj(self) -> CoreV1RuleSpecInputStreamTablesInner:
                    return CoreV1RuleSpecInputStreamTablesInner(
                        name=self.name,
                        watermark=Helpers.maybe(
                            lambda o: o.to_api_obj(), self.watermark
                        ),
                        alias=self.alias,
                        join_type=self.join_type,
                        join_expr=self.join_expr,
                        streaming=self.streaming,
                    )

            tables: Optional[List["Rule.Input.Stream.Table"]] = None
            filter: Optional[str] = None
            sql: Optional[str] = None
            custom: Optional["Rule.Input.CustomStream"] = None
            starting_timestamp: Optional[datetime] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1RuleSpecInputStream],
            ) -> "Rule.Input.Stream":
                if obj is None:
                    return None
                tables = None
                if obj.tables is not None:
                    tables = [
                        Rule.Input.Stream.Table.from_api_obj(item)
                        for item in obj.tables
                    ]

                starting_timestamp = obj.starting_timestamp
                if starting_timestamp is not None and starting_timestamp.tzinfo is None:
                    starting_timestamp = starting_timestamp.replace(tzinfo=timezone.utc)

                return Rule.Input.Stream(
                    tables=tables,
                    filter=obj.filter,
                    sql=obj.sql,
                    custom=Rule.Input.CustomStream.from_api_obj(obj.custom),
                    starting_timestamp=starting_timestamp,
                )

            def to_api_obj(self) -> CoreV1RuleSpecInputStream:
                tables = None
                if self.tables is not None:
                    tables = [item.to_api_obj() for item in self.tables]

                # tzinfo must be attached to the starting timestamp or else
                # the serialization (without trailing time zone) will be
                # rejected by the server.
                starting_timestamp = self.starting_timestamp
                if starting_timestamp is not None and starting_timestamp.tzinfo is None:
                    starting_timestamp = starting_timestamp.replace(tzinfo=timezone.utc)

                return CoreV1RuleSpecInputStream(
                    tables=tables,
                    filter=self.filter,
                    sql=self.sql,
                    custom=Helpers.maybe(lambda o: o.to_api_obj(), self.custom),
                    starting_timestamp=starting_timestamp,
                )

        class Batch(BaseModel):
            """
            Specification for batch input to a Rule, either from a
            SQL statement or a custom notebook.

            Attributes:
                sql (Optional[str]):
                custom (Optional[Rule.Input.CustomBatch]):
            """

            sql: Optional[str] = None
            custom: Optional["Rule.Input.CustomBatch"] = None

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1RuleSpecInputBatch],
            ) -> "Rule.Input.Batch":
                if obj is None:
                    return None
                return Rule.Input.Batch(
                    sql=obj.sql,
                    custom=Rule.Input.CustomBatch.from_api_obj(obj.custom),
                )

            def to_api_obj(self) -> CoreV1RuleSpecInputBatch:
                return CoreV1RuleSpecInputBatch(
                    sql=self.sql,
                    custom=Helpers.maybe(lambda o: o.to_api_obj(), self.custom),
                )

        stream: Optional["Rule.Input.Stream"] = None
        batch: Optional["Rule.Input.Batch"] = None

        @staticmethod
        def from_api_obj(obj: Optional[CoreV1RuleSpecInput]) -> "Rule.Input":
            return Rule.Input(
                stream=Rule.Input.Stream.from_api_obj(obj.stream),
                batch=Rule.Input.Batch.from_api_obj(obj.batch),
            )

        def to_api_obj(self) -> CoreV1RuleSpecInput:
            to_api_obj = lambda o: o.to_api_obj()
            return CoreV1RuleSpecInput(
                stream=Helpers.maybe(to_api_obj, self.stream),
                batch=Helpers.maybe(to_api_obj, self.batch),
            )

    class Observable(BaseModel):
        """
        Observable associated with a Rule.

        Attributes:
            kind (str):
            value (str):
            relationship (str):
            risk (Rule.Observable.Risk):
        """

        class Risk(BaseModel):
            """
            Risk level associated with an Observable.

            Attributes:
                impact (str):
                    A SQL expression indicating the impact (should evaluate to
                    a number between 0-100).
                confidence (str):
                    A SQL expression indicating the confidence (should
                    evaluate to a number between 0-100).
            """

            impact: str
            confidence: str

            @staticmethod
            def from_api_obj(
                obj: Optional[CoreV1RuleObservableRisk],
            ) -> "Rule.Observable.Risk":
                if obj is None:
                    return None
                return Rule.Observable.Risk(
                    impact=obj.impact,
                    confidence=obj.confidence,
                )

            def to_api_obj(self) -> CoreV1RuleObservableRisk:
                return CoreV1RuleObservableRisk(
                    impact=self.impact,
                    confidence=self.confidence,
                )

        kind: str
        value: str
        relationship: str
        risk: "Rule.Observable.Risk"

        @staticmethod
        def from_api_obj(obj: Optional[CoreV1RuleObservable]) -> "Rule.Observable":
            if obj is None:
                return None
            return Rule.Observable(
                kind=obj.kind,
                value=obj.value,
                relationship=obj.relationship,
                risk=Rule.Observable.Risk.from_api_obj(obj.risk),
            )

        def to_api_obj(self) -> CoreV1RuleObservable:
            return CoreV1RuleObservable(
                kind=self.kind,
                value=self.value,
                relationship=self.relationship,
                risk=self.risk.to_api_obj(),
            )

    class Output(BaseModel):
        """
        Output from a Rule after matching.

        Attributes:
            summary (Optional[str]):
            context (Optional[Dict[str, str]]):
            default_context (Optional[bool]):
        """

        summary: Optional[str] = None
        context: Optional[Dict[str, str]] = None
        default_context: Optional[bool] = False

        @staticmethod
        def from_api_obj(obj: Optional[CoreV1RuleSpecOutput]) -> "Rule.Output":
            if obj is None:
                return None
            return Rule.Output(
                summary=obj.summary,
                context=obj.context,
                default_context=obj.default_context,
            )

        def to_api_obj(self) -> CoreV1RuleSpecOutput:
            return CoreV1RuleSpecOutput(
                summary=self.summary,
                context=self.context,
                default_context=self.default_context,
            )

    class Collate(BaseModel):
        """
        Configuration for collating multiple notables generated from a Rule
        within a given timeframe.

        Attributes:
            collate_on (Optional[List[str]]):
                SQL column(s) to collate notables on. This must be either
                'summary', 'context', or a JSON key within 'context'. Only
                one level of nesting is currently supported.
            within (Optional[str]):
                A time duration string for the window to collate within.
            action (Optional[str]):
                append | updateStats | inhibit.
        """

        collate_on: Optional[List[str]] = None
        within: Optional[str] = None
        action: Optional[str] = None

        @staticmethod
        def from_api_obj(obj: Optional[CoreV1RuleSpecCollate]) -> "Rule.Collate":
            if obj is None:
                return None
            return Rule.Collate(
                collate_on=obj.collate_on,
                within=obj.within,
                action=obj.action,
            )

        def to_api_obj(self) -> CoreV1RuleSpecCollate:
            return CoreV1RuleSpecCollate(
                collate_on=self.collate_on,
                within=self.within,
                action=self.action,
            )

    metadata: Optional[Metadata] = None
    rule_metadata: Optional["Rule.RuleMetadata"] = None
    schedule: Schedule
    compute_mode: Optional[str] = None
    input: "Rule.Input"
    observables: Optional[List["Rule.Observable"]] = None
    output: "Rule.Output"
    collate: Optional["Rule.Collate"] = None
    status: Optional[ResourceStatus] = None

    @staticmethod
    def from_api_obj(obj: CoreV1Rule) -> "Rule":
        observables = None
        if obj.spec.observables is not None:
            observables = [
                Rule.Observable.from_api_obj(item) for item in obj.spec.observables
            ]
        return Rule(
            metadata=Metadata.from_api_obj(obj.metadata),
            rule_metadata=Rule.RuleMetadata.from_api_obj(obj.spec.metadata),
            schedule=Schedule.from_api_obj(obj.spec.schedule),
            compute_mode=obj.spec.compute_mode,
            input=Rule.Input.from_api_obj(obj.spec.input),
            observables=observables,
            output=Rule.Output.from_api_obj(obj.spec.output),
            collate=Rule.Collate.from_api_obj(obj.spec.collate),
            status=ResourceStatus.from_api_obj(obj.status),
        )

    @staticmethod
    def from_yaml_str(s: str) -> "Rule":
        docs = yaml.safe_load_all(s)
        docs = list(docs)

        if not docs:
            raise ValueError("YAML is empty")
        if len(docs) > 1:
            raise ValueError(f"Expected a single YAML document, got {len(docs)}")

        data = docs[0]
        if not isinstance(data, dict):
            raise ValueError(
                f"Expected a mapping at top-level, got {type(data).__name__}"
            )

        return Rule.model_validate(data)

    def to_api_obj(self) -> CoreV1Rule:
        observables = None
        if self.observables is not None:
            observables = [item.to_api_obj() for item in self.observables]
        to_api_obj = lambda o: o.to_api_obj()
        return CoreV1Rule(
            api_version="v1",
            kind="Rule",
            metadata=Helpers.maybe(to_api_obj, self.metadata),
            spec=CoreV1RuleSpec(
                metadata=Helpers.maybe(to_api_obj, self.rule_metadata),
                schedule=self.schedule.to_api_obj(),
                compute_mode=self.compute_mode,
                input=self.input.to_api_obj(),
                observables=observables,
                output=self.output.to_api_obj(),
                collate=Helpers.maybe(to_api_obj, self.collate),
            ),
            status=Helpers.maybe(to_api_obj, self.status),
        )



import json
from importlib import resources
from typing import List

_data = json.loads(resources.files(__package__).joinpath("regions.json").read_text())


class Regions:
    @staticmethod
    def lookup(name: str) -> str:
        try:
            return _data[name]
        except KeyError as e:
            raise ValueError(f"unknown region {name}") from e

    @staticmethod
    def list() -> List[str]:
        return list(_data.keys())



from urllib.parse import urlparse
import os


class Helpers:
    default_region = "aws-us-east-1"

    @staticmethod
    def ensure_databricks():
        if "DATABRICKS_RUNTIME_VERSION" not in os.environ:
            raise Exception(
                "attempted to access databricks context outside "
                + "of databricks notebook"
            )

    @staticmethod
    def databricks_context():
        # This import raises an exception if outside a notebook context, so only
        # import if this method is called
        Helpers.ensure_databricks()
        from databricks.sdk.runtime import dbutils

        return dbutils.notebook.entry_point.getDbutils().notebook().getContext()

    @staticmethod
    def current_workspace_url() -> str:
        base_url = Helpers.databricks_context().browserHostName().get()
        return f"https://{base_url}"

    @staticmethod
    def api_token() -> str:
        return Helpers.databricks_context().apiToken().get()

    @staticmethod
    def workspace_name_from_url(url: str) -> str:
        u = urlparse(url)
        return u.hostname
